/*----------------------------------------------------------------------
  File    : napa.c
  Contents: test neurons for assembly participation
  Author  : Christian Borgelt
  History : 2013.02.01 file created from coconad.c
            2013.02.04 first versions of cpc and cxc completed
            2013.02.05 first versions of other tests completed
            2014.03.20 core functions of interval/influence region tests
            2014.03.21 completed cpc, cif and ciw for influence regions
            2014.03.24 bugs in functions austeps() and stepfn() fixed
            2014.03.25 statistics cxc, cir and cwr for influence regions
            2014.03.26 implementation of irs and cws statistics unified
            2014.03.28 time bin based statistics added (bns_...())
            2014.03.31 bugs in functions cws_cif() and cws_ciw() fixed
            2014.04.01 functions bns_cpo() and bns_bre() added
            2014.04.02 point bin based statistics added (pbs_...())
            2014.04.04 optional maximum covering windows version added
            2014.06.27 statistics cif, ciw changed to rate versions
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#ifdef NAPA_MAIN
#ifndef TRN_SURR
#define TRN_SURR
#endif
#ifndef TRN_READ
#define TRN_READ
#endif
#ifndef TA_READ
#define TA_READ
#endif
#endif
#include "train.h"
#ifdef STORAGE
#include "storage.h"
#endif
#ifdef NAPA_MAIN
#include "error.h"
#endif

#ifdef _MSC_VER
#ifndef isnan
#define isnan       _isnan      /* check for 'not a number' */
#endif
#endif                          /* in case C99 is not supported */

#ifndef NAN
#define NAN         (INFINITY-INFINITY)
#endif                          /* not a number */

#define MAXWIN      0           /* whether only maximal windows */

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "napa"
#define DESCRIPTION "test neurons for assembly participation"
#define VERSION     "version 2.7 (2014.10.24)         " \
                    "(c) 2014   Christian Borgelt"

/* --- test statistics --- */
#define STAT_CPC       0        /* conditional pattern complexity */
#define STAT_CXC       1        /* conditional excess complexity */
#define STAT_CCF       2        /* conditional complexity frequency */
#define STAT_CCR       3        /* conditional complexity freq. ratio */
#define STAT_CIF       4        /* conditional item frequency */
#define STAT_CIR       5        /* conditional item freq. ratio */
#define STAT_CIW       6        /* conditional item weight */
#define STAT_CWR       7        /* conditional item weight ratio */
#define STAT_CPO       8        /* conditional pattern overlap */
#define STAT_BRE       9        /* background rate estimation */

/* --- threshold for heap --- */
#define HEAPMIN       15        /* minimum number of trains for heap */

/* --- error codes --- */
/* error codes   0 to  -4 defined in tract.h */
#define E_STDIN      (-5)       /* double assignment of stdin */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* too few/many arguments */
#define E_WIDTH      (-9)       /* invalid influence region width */
#define E_MODE      (-10)       /* invalid statistic mode */
#define E_STAT      (-11)       /* invalid test statistic */
#define E_ALPHA     (-12)       /* invalid weighting exponent */
#define E_SURR      (-13)       /* invalid surrogate gen. method */
#define E_RANDOM    (-14)       /* invalid random function type */
#define E_SIGMA     (-15)       /* invalid random function parameter */
#define E_ITEMS     (-16)       /* too few items found */

#ifndef QUIET                   /* if not quiet version, */
#define MSG         fprintf     /* print messages */
#define XMSG        if (mode & NAPA_VERBOSE) fprintf
#else                           /* if quiet version, */
#define MSG(...)    ((void)0)   /* suppress messages */
#define XMSG(...)   ((void)0)
#endif

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- train heap element --- */
  const double *pt;             /* next point/spike time in train */
  ITEM         item;            /* item/neuron identifier */
} HEAPELEM;                     /* (train heap element) */

typedef struct {                /* --- point process element --- */
  double       pt;              /* point/spike time */
  ITEM         item;            /* item/neuron identifier */
} PPELEM;                       /* (point process element) */

typedef struct {                /* --- step of a step function --- */
  double       pt;              /* point/spike time */
  diff_t       ht;              /* height of step function at point */
} FNSTEP;                       /* (step function step) */

typedef struct {                /* --- time bin statistics data --- */
  TRAINSET     *trnset;         /* underlying train set */
  HEAPELEM     *heap;           /* heap for merging the trains */
  ITEM         **bins;          /* time bin contents  (item sets) */
  ITEM         **sels;          /* selected time bins (item sets) */
  ITEM         *sizes;          /* time bin/item set sizes */
  TID          *bids;           /* time bin identifier list */
  SUPP         *gfs;            /* global      frequencies/weights */
  SUPP         *cfs;            /* conditional frequencies/weights */
  TID          cnt;             /* total       number of time bins */
  TID          cond;            /* conditional number of time bins */
  double       beg;             /* beginning of point/spike range */
  double       prob;            /* prob. of time bin with test item */
  double       avg;             /* average pattern complexity */
} BNSDATA;                      /* (time bin statistics data) */

typedef struct {                /* --- point bin statistics data --- */
  TRAINSET     *trnset;         /* underlying train set */
  HEAPELEM     *heap;           /* heap for merging the trains */
  PPELEM       *pp;             /* trains as a point process */
  ITEM         **bins;          /* point bin contents (item sets) */
  ITEM         *sizes;          /* point bin/item set sizes */
  SUPP         *gfs;            /* global      frequencies/weights */
  SUPP         *cfs;            /* conditional frequencies/weights */
  SUPP         *buf;            /* buffer for computations */
  SUPP         cnt;             /* total       number of points */
  SUPP         cond;            /* conditional number of points */
  double       len;             /* length of point/spike range */
  double       prob;            /* prob. of point bin with test item */
  double       avg;             /* average pattern complexity */
} PBSDATA;                      /* (point bin statistics data) */

typedef struct {                /* --- window statistics data --- */
  TRAINSET     *trnset;         /* underlying train set */
  HEAPELEM     *heap;           /* heap for merging the trains */
  PPELEM       *pp;             /* trains as a point process */
  SUPP         *gfs;            /* global      frequencies/weights */
  SUPP         *cfs;            /* conditional frequencies/weights */
  SUPP         *buf;            /* buffer for computations */
  SUPP         cnt;             /* total       number of windows */
  SUPP         cond;            /* conditional number of windows */
  double       prob;            /* prob. of window with test item */
} CWSDATA;                      /* (covering window statistics data) */

typedef struct {                /* --- infl. reg. statistics data --- */
  TRAINSET     *trnset;         /* underlying train set */
  FNSTEP       *steps;          /* step function of influence regions */
  IVAL         range[2];        /* points/time range plus sentinel */
  IVAL         *ivs1;           /* first  interval buffer (test item) */
  IVAL         *ivs2;           /* second interval buffer (item pair) */
  double       *gls;            /* global      interval lengths */
  double       *cls;            /* conditional interval lengths */
  double       *buf;            /* buffer for item-specific comps. */
  double       len;             /* total length of intervals */
  double       avg;             /* average pattern complexity */
} IRSDATA;                      /* (influence region statistics data) */

typedef double BNSTATFN (BNSDATA *data, TRAIN *train,
                         double width, int alpha);
typedef double PBSTATFN (PBSDATA *data, TRAIN *train,
                         double width, int alpha);
typedef double CWSTATFN (CWSDATA *data, TRAIN *train,
                         double width, int alpha);
typedef double IRSTATFN (IRSDATA *data, TRAIN *train,
                         double width, int alpha);

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#if !defined QUIET && defined NAPA_MAIN
/* --- error messages --- */
static const char *errmsgs[] = {
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_WIDTH    -9 */  "invalid influence region width %g",
  /* E_MODE    -10 */  "invalid statistic mode '%c'",
  /* E_STAT    -11 */  "invalid test statistic '%s'",
  /* E_ALPHA   -12 */  "invalid weighting exponent %d",
  /* E_SURR    -13 */  "invalid surrogate data generation method '%c'",
  /* E_RANDOM  -14 */  "invalid random function type '%c'",
  /* E_SIGMA   -15 */  "invalid random function parameter %g",
  /* E_ITEMS   -16 */  "too few items found (need at least two)",
  /*           -17 */  "unknown error"
};
#endif

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
#ifdef NAPA_MAIN
#ifndef QUIET
static CCHAR    *prgname;       /* program name for error messages */
#endif
static TABREAD  *tread  = NULL; /* table/train reader */
static ITEMBASE *ibase  = NULL; /* item base */
static TRAINSET *trnset = NULL; /* train set */
static RNG      *rng    = NULL; /* random number generator */
static FILE     *out    = NULL; /* output file for test results */
#endif

/*----------------------------------------------------------------------
  Statistics Output Functions
----------------------------------------------------------------------*/

static char sigind (double p)
{                               /* --- get significance indicator */
  if (p < 1e-6) return '#';     /* '#' means significant at 1e-6 */
  if (p < 0.01) return '*';     /* '*' means significant at 1% */
  return (p < 0.05) ? '<' : ' ';/* '<' means significant at 5% */
}  /* sigind() */

/*--------------------------------------------------------------------*/

static void header (FILE *out, long runs, int qtls, const char *fmt)
{                               /* --- print header */
  assert(out && fmt);           /* check the function arguments */
  if (runs <= 0)                /* if no surrogate data */
    fprintf(out, "item     value");
  else {                        /* if to compare to surrogate data */
    fprintf(out, "item     p-value ");
    if (qtls) {                 /* and add point count */
      fprintf(out, fmt, "minimum"); /* minimum */
      fprintf(out, fmt,  "1%");     /* 0.01 quantile */
      fprintf(out, fmt,  "5%");     /* 0.05 quantile */
      fprintf(out, fmt, "25%");     /* 0.25 quantile */
      fprintf(out, fmt, "50%");     /* 0.50 quantile */
      fprintf(out, fmt, "75%");     /* 0.75 quantile */
      fprintf(out, fmt, "95%");     /* 0.95 quantile */
      fprintf(out, fmt, "99%");     /* 0.99 quantile */
      fprintf(out, fmt, "maximum"); /* maximum */
      fprintf(out, fmt, "actual");  /* actual value */
    }
  }
  if (qtls) fprintf(out, " points");  /* add point count and */
  fprintf(out, "\n");           /* terminate the header line */
}  /* header() */

/*--------------------------------------------------------------------*/

static void report (FILE *out, const char *fmt,
                    double stats[], long runs, double act, size_t n)
{                               /* --- report statistics information */
  assert(out && fmt && stats);  /* check the function arguments */
  fprintf(out, fmt, stats[0]);      /* minimum */
  fprintf(out, fmt, stats[(int)(0.01*(double)runs)]);/* 0.01 quantile */
  fprintf(out, fmt, stats[(int)(0.05*(double)runs)]);/* 0.05 quantile */
  fprintf(out, fmt, stats[(int)(0.25*(double)runs)]);/* 0.25 quantile */
  fprintf(out, fmt, stats[(int)(0.50*(double)runs)]);/* 0.50 quantile */
  fprintf(out, fmt, stats[(int)(0.75*(double)runs)]);/* 0.75 quantile */
  fprintf(out, fmt, stats[(int)(0.95*(double)runs)]);/* 0.95 quantile */
  fprintf(out, fmt, stats[(int)(0.99*(double)runs)]);/* 0.99 quantile */
  fprintf(out, fmt, stats[runs-1]); /* maximum */
  fprintf(out, fmt, act);           /* actual value */
  fprintf(out, " %6"SIZE_FMT, n);   /* number of points */
}  /* report() */

/*--------------------------------------------------------------------*/

static void footer (FILE *out, double sums[], long runs,
                    double total, size_t points, int qtls)
{
  long   k;                     /* loop variable for runs */
  double p;                     /* computed p-value */

  assert(out && sums && (runs > 0));  /* check function arguments */
  dbl_qsort(sums, (size_t)runs, +1);  /* sort summed statistics */
  for (p = 0, k = 0; k < runs; k++)
    if (sums[k] >= total) p++;  /* compute the p-value as the number */
  p /= (double)runs;            /* of runs at least the total value */
  printf("total   %8.6f%c", p, sigind(p));
  if (qtls)                     /* if extended output requested */
    report(out, " %11.4f", sums, runs, total, points);
  fprintf(out, "\n");           /* report quantiles of statistic */
}  /* footer() */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static double powi (double x, int n)
{                               /* --- raise x to the power of n */
  double r = (n & 1) ? x : 1;   /* result */

  assert(n >= 0);               /* check the function arguments */
  for (n >>= 1; n > 0; n >>= 1){/* traverse the powers of 2 and */
    x *= x; if (n & 1) r *= x; }/* multiply them into the result */
  return r;                     /* if the corr. exponent bit is set, */
}  /* powi() */                 /* finally return the result */

/*--------------------------------------------------------------------*/

static diff_t ipowi (diff_t x, int n)
{                               /* --- raise x to the power of n */
  diff_t r = (n & 1) ? x : 1;   /* result */

  assert(n >= 0);               /* check the function arguments */
  for (n >>= 1; n > 0; n >>= 1){/* traverse the powers of 2 and */
    x *= x; if (n & 1) r *= x; }/* multiply them into the result */
  return r;                     /* if the corr. exponent bit is set, */
}  /* ipowi() */                /* finally return the result */

/*--------------------------------------------------------------------*/

static void sift (HEAPELEM heap[], ITEM lft, ITEM rgt)
{                               /* --- let element sift down in heap */
  ITEM     i;                   /* index of first successor in heap */
  double   x;                   /* buffer for successor point */
  HEAPELEM t;                   /* buffer for element to sift down */

  assert(heap && (lft >= 0) && (rgt > lft));
  t = heap[lft];                /* note sift element */
  i = lft +lft +1;              /* compute index of first successor */
  do {                          /* sift loop */
    x = *heap[i].pt;            /* get point of first successor */
#if 1                           /* ignore item for sorting */
    if ((i             < rgt)   /* if second successor exists and */
    &&  (*heap[i+1].pt < x))    /* it is smaller than the first, */
      x = *heap[++i].pt;        /* go to the second successor */
    if (x > *t.pt)              /* if the successor is no less */
      break;                    /* than the sift element, abort */
#else                           /* respect item for equal points */
    if ((i             < rgt)   /* if second successor exists and */
    && (  (*heap[i+1].pt <  x)  /* it is smaller than the first, */
    ||  (((*heap[i+1].pt == x) && (heap[i+1].item < heap[i].item)))))
      x = *heap[++i].pt;        /* go to the second successor */
    if ((x >  *t.pt)            /* if the successor is no less */
    || ((x == *t.pt) && (heap[i].item >= t.item)))
      break;                    /* than the sift element, abort */
#endif
    heap[lft] = heap[i];        /* let the successor ascend in heap */
    lft = i;                    /* note index of current element */
    i  += i+1;                  /* compute index of first successor */
  } while (i <= rgt);           /* while still within heap */
  heap[lft] = t;                /* store the sift element */
}  /* sift() */

/*--------------------------------------------------------------------*/

static int tns2pp (TRAINSET *trnset, PPELEM pp[], HEAPELEM heap[])
{                               /* --- train set to point process */
  ITEM         i, k, n;         /* loop variables */
  TRAIN        *train;          /* to traverse the trains */
  const double *t;              /* to traverse the points */
  double       min;             /* minimum point/spike */
  HEAPELEM     *h = NULL;       /* created heap if not given */

  assert(trnset && pp);         /* check the function arguments */
  n = tns_cnt(trnset);          /* get the number of trains */
  if (n <= 1) {                 /* if there is only one train */
    train = tns_train(trnset, 0);
    for (t = trn_points(train); *t < TRN_END; t++) {
      pp->pt = *t; pp->item = 0; pp++; }
    pp->pt   = TRN_END;         /* copy the points of the only train */
    pp->item = -1; return 0;    /* and store a sentinel element */
  }                             /* at the end of the point process */
  if (!heap) {                  /* if there is no heap given */
    heap = h = (HEAPELEM*)malloc((size_t)n *sizeof(HEAPELEM));
    if (!heap) return -1;       /* create internal heap for merging */
  }                             /* which is deleted in the end */
  for (i = 0; i < n; i++) {     /* traverse the trains */
    heap[i].pt   = trn_points(tns_train(trnset, i));
    heap[i].item = i;           /* get point arrays of all trains */
  }                             /* and their corresponding items */
  if (n < HEAPMIN) {            /* if there are only few trains */
    while (1) {                 /* minimum point selection loop */
      for (min = *heap[i = k = 0].pt; ++i < n; )
        if (*heap[i].pt < min)  /* find the train */
          min = *heap[k = i].pt;/* with the minimum point */
      if (min >= TRN_END) break;/* if all trains are exhausted, abort */
      pp->pt   = min;           /* store the minimum point */
      pp->item = heap[k].item;  /* and the corresponding item */
      pp++;      heap[k].pt++;  /* consume the processed point */
    } }
  else {                        /* if there are many items */
    for (i = n-- >> 1; --i >= 0; )
      sift(heap, i, n);         /* build the point heap */
    while (1) {                 /* minimum point selection loop */
      min = *heap[0].pt;        /* get the minimum point */
      if (min >= TRN_END) break;/* if all trains are exhausted, abort */
      pp->pt   = min;           /* store the minimum point */
      pp->item = heap[0].item;  /* and the corresponding item */
      pp++;      heap[0].pt++;  /* consume the processed point */
      sift(heap, 0, n);         /* let the new point sift down */
    }                           /* in heap (restore heap structure) */
  }
  pp->pt   = TRN_END;           /* store a sentinel element */
  pp->item = -1;                /* at the end of the point process */
  if (h) free(h);               /* delete an internal heap */
  return 0;                     /* return 'ok' */
}  /* tns2pp() */

/*--------------------------------------------------------------------*/

static void ppxtrn (PPELEM pp[], ITEM item, TRAIN *train)
{                               /* --- modify a point process */
  PPELEM *src, *dst;            /* destination for merging */
  const double *t;              /* to traverse points of the train */

  assert(pp && (train || (item >= 0))); /* check function arguments */
  if (item < 0)                 /* if no item is explicitely given, */
    item = trn_item(train);     /* get the item of the train */
  for (dst = src = pp; src->pt < TRN_END; src++)
    if (src->item != item)      /* remove all points from the process */
      *dst++ = *src;            /* that are associated with the item */
  if (!train) {                 /* if only to remove item, abort */
    dst->pt = TRN_END; dst->item = -1; return; }
  assert((SUPP)(src-dst) == trn_len(train));
  while (dst > pp)              /* move remaining points */
    *--src = *--dst;            /* to the end of the point process */
  for (t = trn_points(train); 1; dst++) {
    if      (*t < src->pt) { dst->pt = *t++; dst->item = item; }
    else if (*t > src->pt) *dst = *src++;
    else if (*t < TRN_END) *dst = *src++;
    else break;                 /* merge point process and train */
  }                             /* to produce a modified process */
  assert(dst->pt == TRN_END);   /* check the number of points again */
  dst->pt   = TRN_END;          /* store a sentinel element */
  dst->item = -1;               /* at the end of the point process */
}  /* ppxtrn() */

/*----------------------------------------------------------------------
  Time Bin Statistics
----------------------------------------------------------------------*/

static void bns_free (BNSDATA *data)
{                               /* --- free statistics data */
  assert(data);                 /* check the function argument */
  if (data->sels)  free(data->sels);
  if (data->bids)  free(data->bids);
  if (data->heap)  free(data->heap);
  free(data->sizes);
  free(data->bins);
}  /* bns_free() */

/*--------------------------------------------------------------------*/

static int bns_alloc (BNSDATA *data, TRAINSET *trnset, int stat,
                      double width, double beg, double end)
{                               /* --- allocate statistics data */
  ITEM     i, m, n;             /* loop variable, number of items */
  TID      k, p;                /* time bin index/identifier */
  size_t   z;                   /* size of a buffer element */
  HEAPELEM *heap;               /* heap for merging the spike trains */
  ITEM     *bin;                /* to traverse the time bins */
  double   min;                 /* minimum point/spike time */

  assert(data && trnset         /* check the function arguments */
  &&    (width > 0) && !isnan(beg) && !isnan(end) && (end > beg));
  data->trnset = trnset;        /* note the underlying train set */
  data->beg    = beg;           /* and the point/spike time start */
  data->heap   = NULL;          /* clear fields for easy clean-up */
  data->bids   = NULL;          /* (allows to call bns_free()) */
  data->sels   = NULL;          /* compute number of time bins */
  data->cnt    = (TID)ceil((end-beg) /width);
  data->bins   = (ITEM**)malloc(((size_t)data->cnt+1) *sizeof(ITEM*));
  if (!data->bins) return -1;   /* allocate time bin array */
  z = tns_extent(trnset) +2 *(size_t)data->cnt;
  data->sizes = bin = (ITEM*)malloc(z *sizeof(ITEM));
  if (!bin) { bns_free(data); return -1; }
  bin += data->cnt;             /* allocate size array and bins */
  n = tns_cnt(trnset);          /* get the number of items */
  z = 2*sizeof(SUPP);           /* create a heap for train merging */
  if (sizeof(HEAPELEM) > z) z = sizeof(HEAPELEM);
  data->heap = heap = (HEAPELEM*)malloc((size_t)(n+1) *z);
  if (!heap) { bns_free(data); return -1; }
  data->gfs  = (SUPP*)data->heap;         /* use it also as a buffer */
  data->cfs  = data->gfs +n+1;  /* for global and conditional freqs. */
  data->bids = (TID*) malloc((size_t)data->cnt *sizeof(TID));
  if (!data->bids) { bns_free(data); return -1; }
  if (stat == STAT_CPO) {       /* if conditional pattern overlap */
    data->sels = (ITEM**)malloc((size_t)data->cnt *sizeof(ITEM*));
    if (!data->sels) { bns_free(data); return -1; }
  }                             /* allocate time bin selection array */
  for (i = 0; i < n; i++) {     /* traverse the trains */
    data->heap[i].pt   = trn_points(tns_train(trnset, i));
    data->heap[i].item = i;     /* get point arrays of all trains */
  }                             /* and their corresponding items */
  data->bins[p = 0] = bin;      /* start with the first time bin */
  if (n < HEAPMIN) {            /* if there are only few trains */
    while (1) {                 /* minimum point selection loop */
      for (min = *heap[i = m = 0].pt; ++i < n; )
        if (*heap[i].pt < min)  /* find the train */
          min = *heap[m = i].pt;/* with the minimum point */
      if (min >= TRN_END) break;/* if all trains are exhausted, abort */
      k = (TID)floor((min -beg) /width);
      if (k >= data->cnt) k = data->cnt-1;
      while (p < k) {           /* while beyond current time bin */
        *bin++ = TA_END;        /* store a sentinel after the items */
        data->bins[++p] = bin;  /* start the next time bin */
      }                         /* (this may produce empty bins) */
      *bin++ = heap[m].item;    /* store the next item and */
      heap[m].pt++;             /* skip the processed point */
    } }
  else {                        /* if there are many items */
    for (i = n-- >> 1; --i >= 0; )
      sift(heap, i, n);         /* build the point heap */
    while (1) {                 /* minimum point selection loop */
      min = *heap[0].pt;        /* get the minimum point */
      if (min >= TRN_END) break;/* if all trains are exhausted, abort */
      k = (TID)floor((min -beg) /width);
      if (k >= data->cnt) k = data->cnt-1;
      while (p < k) {           /* while beyond current time bin */
        *bin++ = TA_END;        /* store a sentinel after the items */
        data->bins[++p] = bin;  /* start the next time bin */
      }                         /* (this may produce empty bins) */
      *bin++ = heap[0].item;    /* store the next item */
      heap[0].pt++;             /* skip the processed point and */
      sift(heap, 0, n);         /* let the new point sift down */
    }                           /* in heap (restore heap structure) */
  }
  while (p < data->cnt) {       /* while not at last time bin */
    *bin++ = TA_END;            /* store a sentinel after the items */
    data->bins[++p] = bin;      /* start the next time bin */
  }                             /* (this may produce empty bins) */
  for (k = 0; k < data->cnt; k++) { /* traverse the time bins */
    m = (ITEM)(data->bins[k+1] -data->bins[k] -1);
    ia_qsort(data->bins[k], (size_t)m, +1);
    data->sizes[k] = m = ia_unique(data->bins[k], (size_t)m);
    data->bins[k][m] = TA_END;  /* remove duplicate items from bins */
  }                             /* and store a sentinel at the end */
  return 0;                     /* return 'ok' */
}  /* bns_alloc() */

/*--------------------------------------------------------------------*/

static void bns_init (BNSDATA *data, ITEM item,
                      double width, int stat, int alpha)
{                               /* --- initialize statistics data */
  TID  k, p;                    /* time bin index/identifier */
  TID  *b;                      /* to traverse a bin id list */
  ITEM *s;                      /* to traverse items of a time bin */
  ITEM max;                     /* maximum number of items for bre */
  const double *t;              /* to traverse the train */

  assert(data                   /* check the function arguments */
  &&    (item >= 0) && (item < tns_cnt(data->trnset))
  &&    (width > 0) && (alpha > 0));
  b = data->bids; p = -1;       /* get the time bin identifier list */
  for (t = trn_points(tns_train(data->trnset,item)); *t < TRN_END; t++){
    k = (TID)floor((*t -data->beg) /width);
    if (k <  0)         k = 0;  /* compute time bin of point */
    if (k >= data->cnt) k = data->cnt-1;
    if (k == p) continue;       /* check and adapt time bin index */
    data->sizes[*b++ = p = k]--;/* remove item from time bin and */
  }                             /* collect the time bin identifiers */
  *b = -1;                      /* store a sentinel at the end */
  memset(data->gfs, 0, ((size_t)tns_cnt(data->trnset)+1) *sizeof(SUPP));
  data->avg = 0;                /* clear global frequencies/weights */
  if (data->cnt <= 0) return;   /* if there are no time bins, abort */
  if ((stat == STAT_CPC) || (stat == STAT_CXC)
  ||  (stat == STAT_CCF) || (stat == STAT_CCR)) {
    if (stat != STAT_CPC) {     /* if not cond. pattern complexity */
      for (k = 0; k < data->cnt; k++)
        data->avg += (double)data->sizes[k]; }
    else {                      /* if conditional pattern complexity */
      for (k = 0; k < data->cnt; k++)
        data->avg += powi((double)data->sizes[k], alpha);
    }                           /* include the exponent alpha */
    data->avg /= (double)data->cnt;
  }                             /* compute average pattern complexity */
  if ((stat == STAT_CCF) || (stat == STAT_CCR)) {
    for (k = 0; k < data->cnt; k++)
      data->gfs[data->sizes[k]] += 1;
  }                             /* collect global complexity freqs. */
  if ((stat == STAT_CIF) || (stat == STAT_CIR)) {
    for (k = 0; k < data->cnt; k++)
      for (s = data->bins[k]; *s != TA_END; s++)
        data->gfs[*s] += 1;     /* collect global item frequencies */
    data->gfs[item] = 0;        /* and clear frequency for test item */
  }
  if ((stat == STAT_CIW) || (stat == STAT_CWR)) {
    for (k = 0; k < data->cnt; k++)
      for (s = data->bins[k]; *s != TA_END; s++)
        data->gfs[*s] += data->sizes[k];
    data->gfs[item] = 0;        /* collect global item weights */
  }                             /* and clear weight for test item */
  if (stat == STAT_BRE) {       /* if background rate estimation */
    for (max = (ITEM)(alpha-1), k = 0; k < data->cnt; k++)
      if (data->sizes[k] <= max) data->avg += 1;
  }                             /* count bins with <= alpha points */
}  /* bns_init() */

/*--------------------------------------------------------------------*/

static void bns_term (BNSDATA *data, ITEM item)
{                               /* --- terminate computations */
  TID *b;                       /* to traverse a bin id list */

  assert(data                   /* check the function arguments */
  &&    (item >= 0) && (item < tns_cnt(data->trnset)));
  for (b = data->bids; *b >= 0; b++)
    data->sizes[*b]++;          /* re-add item to complexities */
}  /* bns_term() */

/*--------------------------------------------------------------------*/

static void bns_cxfrq (BNSDATA *data, TRAIN *train, double width)
{                               /* --- collect complexity frequencies */
  TID  k, p, n;                 /* time bin index/identifier, counter */
  const double *t;              /* to traverse points of the train */

  assert(data && train && (width >= 0)); /* check function arguments */
  memset(data->cfs, 0, ((size_t)tns_cnt(data->trnset)+1) *sizeof(SUPP));
  p = -1; n = 0;                /* clear the conditional frequencies */
  for (t = trn_points(train); *t < TRN_END; t++) {
    k = (TID)floor((*t -data->beg) /width);
    if (k <  0)         k = 0;  /* compute time bin of point */
    if (k >= data->cnt) k = data->cnt-1;
    if (k == p) continue;       /* check and adapt time bin index */
    data->cfs[data->sizes[p = k]] += 1;
    n += 1;                     /* count the conditional complexities */
  }                             /* and count the time bins */
  data->prob = (double)n/(double)data->cnt;
}  /* bns_cxfrq() */            /* compute item occurrence prob. */

/*--------------------------------------------------------------------*/

static void bns_itfrq (BNSDATA *data, TRAIN *train, double width)
{                               /* --- collect item frequencies */
  TID  k, p, n;                 /* time bin index/identifier, counter */
  ITEM *s;                      /* to traverse items of a time bin */
  const double *t;              /* to traverse points of the train */

  assert(data && train && (width >= 0)); /* check function arguments */
  memset(data->cfs, 0, ((size_t)tns_cnt(data->trnset)+1) *sizeof(SUPP));
  p = -1; n = 0;                /* clear the conditional frequencies */
  for (t = trn_points(train); *t < TRN_END; t++) {
    k = (TID)floor((*t -data->beg) /width);
    if (k <  0)         k = 0;  /* compute time bin of point */
    if (k >= data->cnt) k = data->cnt-1;
    if (k == p) continue;       /* check and adapt time bin index */
    for (s = data->bins[p = k]; *s != TA_END; s++)
      data->cfs[*s] += 1;       /* count conditional item frequencies */
    n += 1;                     /* and count the time bins, but */
  }                             /* clear frequency for test item */
  data->cfs[trn_item(train)] = 0;
  data->cond = n;               /* compute item occurrence prob. */
  data->prob = (double)n /(double)data->cnt;
}  /* bns_itfrq() */

/*--------------------------------------------------------------------*/

static void bns_itwgt (BNSDATA *data, TRAIN *train, double width)
{                               /* --- collect item frequencies */
  TID  k, p, n;                 /* time bin index/identifier, counter */
  ITEM *s;                      /* to traverse items of a time bin */
  const double *t;              /* to traverse the train */

  assert(data && train && (width >= 0)); /* check function arguments */
  memset(data->cfs, 0, ((size_t)tns_cnt(data->trnset)+1) *sizeof(SUPP));
  p = -1; n = 0;                /* init. previous bin id and counter */
  for (t = trn_points(train); *t < TRN_END; t++) {
    k = (TID)floor((*t -data->beg) /width);
    if (k <  0)         k = 0;  /* compute time bin of point */
    if (k >= data->cnt) k = data->cnt-1;
    if (k == p) continue;       /* check and adapt time bin index */
    for (s = data->bins[p = k]; *s != TA_END; s++)
      data->cfs[*s] += data->sizes[k];
    n += 1;                     /* sum conditional item weights */
  }                             /* and count the time bins, but */
  data->cfs[trn_item(train)] = 0; /* clear weight for test item */
  data->cond = n;               /* compute item occurrence prob. */
  data->prob = (double)n /(double)data->cnt;
}  /* bns_itwgt() */

/*--------------------------------------------------------------------*/

static double bns_cpc (BNSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional pattern complexity */
  TID    k, p, n;               /* time bin index/identifier, counter */
  double sum;                   /* sum of pattern complexities */
  const double *t;              /* to traverse points of the train */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  sum = 0; n = 0; p = -1;       /* initialize the statistic */
  for (t = trn_points(train); *t < TRN_END; t++) {
    k = (TID)floor((*t -data->beg) /width);
    if (k <  0)         k = 0;  /* compute time bin of point */
    if (k >= data->cnt) k = data->cnt-1;
    if (k == p) continue;       /* check and adapt time bin index */
    sum += powi((double)data->sizes[p = k], alpha);
    n   += 1;                   /* sum the pattern complexities */
  }                             /* and count the time bins */
  if (n > 0) sum /= (double)n;  /* normalize the pattern complexity */
  return (data->avg > 0) ? (sum -data->avg) /data->avg : 0;
}  /* bns_cpc() */

/*--------------------------------------------------------------------*/

static double bns_cxc (BNSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional excess complexity */
  TID    k, p, n;               /* time bin index/identifier, counter */
  double x, sum;                /* (sum of) excess complexities */
  const double *t;              /* to traverse the train */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  sum = 0; n = 0; p = -1;       /* initialize the statistic */
  for (t = trn_points(train); *t < TRN_END; t++) {
    k = (TID)floor((*t -data->beg) /width);
    if (k <  0)         k = 0;  /* compute time bin of point */
    if (k >= data->cnt) k = data->cnt-1;
    if (k == p) continue;       /* check and adapt time bin index */
    x = (double)data->sizes[p = k] -data->avg;
    if (x > 0) sum += powi(x, alpha);
    n += 1;                     /* sum the excess complexities */
  }                             /* and count the time bins */
  return (n > 0) ? sum /(double)n : 0;
}  /* bns_cxc() */              /* finally normalize the statistic */

/*--------------------------------------------------------------------*/

static double bns_ccf (BNSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional complexity freq. */
  ITEM   i, n;                  /* loop variable, number of items */
  double sum, x;                /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  bns_cxfrq(data, train, width);/* compute the complexity frequencies */
  n = tns_cnt(data->trnset);    /* traverse the pattern complexities */
  for (sum = 0, i = (int)ceil(data->avg); i <= n; i++) {
    x = (double)data->cfs[i] -data->prob *(double)data->gfs[i];
    if (x > 0) sum += x *powi((double)i -data->avg, alpha);
  }                             /* sum the excess frequencies */
  return (n > 1) ? sum /(double)n : 0;
}  /* bns_ccf() */              /* finally normalize the statistic */

/*--------------------------------------------------------------------*/

static double bns_ccr (BNSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- cond. complexity freq. ratio */
  ITEM   i, n;                  /* loop variable, number of items */
  double sum, x, y;             /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  bns_cxfrq(data, train, width);/* compute the complexity frequencies */
  n = tns_cnt(data->trnset);    /* traverse the pattern complexities */
  for (sum = 0, i = (int)ceil(data->avg); i <= n; i++) {
    x = (double)data->cfs[i] -(y = data->prob *(double)data->gfs[i]);
    if (x > 0) sum += x/(y+1) *powi((double)i -data->avg, alpha);
  }                             /* sum the excess frequencies ratios */
  return (n > 1) ? sum /(double)n : 0;
}  /* bns_ccr() */              /* finally normalize the statistic */

/*--------------------------------------------------------------------*/

static double bns_cif (BNSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional item frequency */
  ITEM   i, n;                  /* loop variable, number of items */
  double sum, x;                /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  bns_itfrq(data, train, width);/* compute item frequencies */
  n = tns_cnt(data->trnset);    /* and traverse the items */
  for (sum = 0, i = 0; i < n; i++) {
    #if 0
    x = (double)data->cfs[i] -data->prob *(double)data->gfs[i];
    #else
    x = (double)data->cfs[i] /(double)data->cond
      - (double)data->gfs[i] /(double)data->cnt;
    #endif
    if (x > 0) sum += powi(x, alpha);
  }                             /* sum the excess item frequencies */
  return (n > 1) ? sum /(double)n : 0;
}  /* bns_cif() */              /* finally normalize the statistic */

/*--------------------------------------------------------------------*/

static double bns_cir (BNSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional item freq. ratio */
  ITEM   i, n;                  /* loop variable, number of items */
  double sum, x, y;             /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  bns_itfrq(data, train, width);/* compute item frequencies */
  n = tns_cnt(data->trnset);    /* and traverse the items */
  for (sum = 0, i = 0; i < n; i++) {
    #if 0
    x = (double)data->cfs[i] -(y = data->prob *(double)data->gfs[i]);
    #else
    y = (double)data->gfs[i] /(double)data->cnt;
    x = (double)data->cfs[i] /(double)data->cond -y;
    #endif
    if (x > 0) sum += powi(x/y, alpha);
  }                             /* sum excess item frequencies ratios */
  return (n > 1) ? sum /(double)n : 0;
}  /* bns_cir() */              /* finally normalize the statistic */

/*--------------------------------------------------------------------*/

static double bns_ciw (BNSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional item weight */
  ITEM   i, n;                  /* loop variable, number of items */
  double sum, x;                /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  bns_itwgt(data, train, width);/* compute the item weights */
  n = tns_cnt(data->trnset);    /* and traverse the items */
  for (sum = 0, i = 0; i < n; i++) {
    #if 0
    x = (double)data->cfs[i] -data->prob *(double)data->gfs[i];
    #else
    x = (double)data->cfs[i] /(double)data->cond
      - (double)data->gfs[i] /(double)data->cnt;
    #endif
    if (x > 0) sum += powi(x, alpha);
  }                             /* sum the excess weights */
  return (n > 1) ? sum /(double)n : 0;
}  /* bns_ciw() */              /* finally normalize the statistic */

/*--------------------------------------------------------------------*/

static double bns_cwr (BNSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional item weight ratio */
  ITEM   i, n;                  /* loop variable, number of items */
  double sum, x, y;             /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  bns_itwgt(data, train, width);/* compute the item weights */
  n = tns_cnt(data->trnset);    /* and traverse the items */
  for (sum = 0, i = 0; i < n; i++) {
    #if 0
    x = (double)data->cfs[i] -(y = data->prob *(double)data->gfs[i]);
    #else
    y = (double)data->gfs[i] /(double)data->cnt;
    x = (double)data->cfs[i] /(double)data->cond -y;
    #endif
    if (x > 0) sum += powi(x/y, alpha);
  }                             /* sum the excess weight ratios */
  return (n > 1) ? sum /(double)n : 0;
}  /* bns_cwr() */              /* finally normalize the statistic */

/*--------------------------------------------------------------------*/

static double bns_cpo (BNSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional pattern overlap */
  TID    k, p, n;               /* time bin index/identifier, counter */
  ITEM   i, m;                  /* item, number of items in overlap */
  ITEM   *s;                    /* to traverse items of a time bin */
  double sum;                   /* sum of pattern overlaps */
  const double *t;              /* to traverse points of the train */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  n = 0; p = -1;                /* initialize the time bin counter */
  for (t = trn_points(train); *t < TRN_END; t++) {
    k = (TID)floor((*t -data->beg) /width);
    if (k <  0)         k = 0;  /* compute time bin of point */
    if (k >= data->cnt) k = data->cnt-1;
    if (k == p) continue;       /* check and adapt time bin index */
    data->sels[n++] = data->bins[p = k];
  }                             /* collect the selected time bins */
  if (n < 2) return 0;          /* check for at least two time bins */
  memset(data->cfs, 0, (size_t)tns_cnt(data->trnset) *sizeof(SUPP));
  sum = 0;                      /* initialize sum and item flags */
  i   = trn_item(train);        /* get the item to test and */
  for (k = 1; k < n; k++) {     /* traverse the time bins */
    for (s = data->sels[k]; *s != TA_END; s++)
      data->cfs[*s] = 1;        /* set flags for contained items */
    data->cfs[i] = 0;           /* but exclude the item to test */
    for (p = k; --p >= 0; ) {   /* traverse preceding time bins */
      for (m = 0, s = data->sels[p]; *s != TA_END; s++)
        m += data->cfs[*s];     /* count items in time bin overlap */
      if (m > 1) sum += powi((double)m, alpha);
    }                           /* sum the pattern overlaps */
    for (s = data->sels[k]; *s != TA_END; s++)
      data->cfs[*s] = 0;        /* clear flags for contained items */
  }                             /* finally normalize summed overlaps */
  return 0.5*sum /((double)n*(double)(n-1));
}  /* bns_cpo() */

/*--------------------------------------------------------------------*/

static double bns_bre (BNSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional pattern complexity */
  TID    k, p, n;               /* time bin index/identifier, counter */
  double sum;                   /* sum of pattern complexities */
  double eta, xi;               /* probabilities */
  const double *t;              /* to traverse points of the train */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  alpha -= 1;                   /* compute maximum number of items */
  sum = n = 0; p = -1;          /* initialize the statistic */
  for (t = trn_points(train); *t < TRN_END; t++) {
    k = (TID)floor((*t -data->beg) /width);
    if (k <  0)         k = 0;  /* compute time bin of point */
    if (k >= data->cnt) k = data->cnt-1;
    if (k == p) continue;       /* check and adapt time bin index */
    if (data->sizes[p = k] <= alpha) sum += 1;
    n += 1;                     /* sum the pattern complexities */
  }                             /* and count the time bins */
  eta = (double)n /(double)data->cnt;
  xi  = sum /data->avg;         /* marginal and background prob. */
  xi  = (xi < 1) ? (eta -xi) /(1.0 -xi) : 0;
  xi /= eta;                    /* compute (rel.) firing probability */
  return xi;                    /* return the value of the statistic */
}  /* bns_bre() */

/*--------------------------------------------------------------------*/

static int bns_test (TRAINSET *trnset, ITEM item,
                     double width, double beg, double end,
                     int surr, RNG *rng, int stat, int alpha,
                     long runs, int qtls, FILE *out)
{                               /* --- execute statistical test */
  int       r = 0;              /* function result, error status */
  ITEM      i, n, e;            /* loop variables, buffer */
  long      k;                  /* loop variable for runs */
  size_t    z, points;          /* (total) number of points/spikes */
  TRAIN     *train, *t;         /* surrogate and original train */
  TRNSURRFN *surrfn;            /* surrogate train generation func. */
  BNSTATFN  *bnsfn;             /* test statistic to compute */
  double    *stats = NULL;      /* train statistic values */
  double    *sums  = NULL;      /* sums of statistic values */
  double    x, y, total;        /* (total value of) test statistic */
  double    p;                  /* computed p-value */
  CCHAR     *s;                 /* name/identifier of an item/neuron */
  BNSDATA   data;               /* influence region statistics data */

  assert(trnset && (width >= 0) && (end >= beg) && rng
  &&    (alpha > 0) && out);    /* check the function arguments */

  /* --- initialize --- */
  train = trn_create(0, NULL, tns_maxlen(trnset));
  if (!train) return -1;        /* create a train for the surrogates */
  if (runs > 0) {               /* if to generate surrogates */
    stats = (double*)malloc((size_t)(runs +runs) *sizeof(double));
    if (!stats) { trn_delete(train); return -1; }
    sums = stats +runs;         /* allocate the statistics arrays */
    dbl_clear(sums, (size_t)runs);
  }                             /* clear the statistic sums */
  total = 0; points = 0;        /* initialize the statistic total */
  if (bns_alloc(&data, trnset, stat, width, beg, end) != 0) {
    if (stats) { free(stats); } trn_delete(train); return -1; }
  if      (surr == 'r')      surrfn = trn_random;
  else if (surr == 'd')      surrfn = trn_dither;
  else if (surr == 's')      surrfn = trn_shift;
  else                       surrfn = trn_random;
  if      (stat == STAT_CPC) bnsfn  = bns_cpc;
  else if (stat == STAT_CXC) bnsfn  = bns_cxc;
  else if (stat == STAT_CCF) bnsfn  = bns_ccf;
  else if (stat == STAT_CCR) bnsfn  = bns_ccr;
  else if (stat == STAT_CIF) bnsfn  = bns_cif;
  else if (stat == STAT_CIR) bnsfn  = bns_cir;
  else if (stat == STAT_CIW) bnsfn  = bns_ciw;
  else if (stat == STAT_CWR) bnsfn  = bns_cwr;
  else if (stat == STAT_CPO) bnsfn  = bns_cpo;
  else if (stat == STAT_BRE) bnsfn  = bns_bre;
  else                       bnsfn  = bns_cpc;
  header(out, runs, qtls, " %11s");

  /* --- execute tests for items --- */
  n = tns_cnt(trnset);          /* get the number of trains */
  if (item >= 0) e = (i = item)+1;
  else         { e = n; i = 0; }/* get the range of items */
  for ( ; i < e; i++) {         /* traverse the items to analyze */
    t = tns_train(trnset, i);   /* sum the number of points */
    points += z = (size_t)trn_len(t);
    bns_init(&data, i, width, stat, alpha);
    x = (z > 0) ? bnsfn(&data, t, width, alpha) : 0;
    total += x;                 /* compute and sum the test statistic */
    s = ib_name(tns_base(trnset), i);
    if (runs <= 0) {            /* if no surrogate data requested, */
      fprintf(out, "%6s  %11.6f", s, x); /* report statistic value */
      if (qtls) fprintf(out, " %6"SIZE_FMT, z); }
    else {                      /* if to compare to surrogate data */
      for (p = 0, k = runs; --k >= 0; ) {
        train = surrfn(tns_train(trnset, i), beg, end, rng, train);
        trn_sort(train, +1);    /* create a surrogate train */
        y = (z > 0) ? bnsfn(&data, train, width, alpha) : 0;
        stats[k] = y;           /* compute and store test statistic */
        sums[k] += y;           /* sum test statistic over items */
        if (y >= x) p++;        /* count the values exceeding or */
      }                         /* meeting the actual statistic value */
      p /= (double)runs;        /* compute and print the p-value */
      fprintf(out, "%6s  %8.6f%c", s, p, sigind(p));
      if (qtls) {               /* if extended output requested */
        dbl_qsort(stats, (size_t)runs, +1);
        report(out, " %11.6f", stats, runs, x, z);
      }                         /* sort statistic values and */
    }                           /* report quantiles of statistic */
    fprintf(out, "\n");         /* and the actual statistic value */
    bns_term(&data, i);         /* terminate the output line */
  }                             /* and the statistic computation */

  /* --- total statistics --- */
  if ((runs > 0) && (item < 0)) /* if all neurons were checked */
    footer(out, sums, runs, total, points, qtls);

  /* --- clean up --- */
  bns_free(&data);              /* delete the work memory, */
  if (stats) free(stats);       /* the statistics arrays */
  trn_delete(train);            /* and the surrogate train */
  return r;                     /* return the error status */
}  /* bns_test() */

/*----------------------------------------------------------------------
  Point Bin Statistics
----------------------------------------------------------------------*/

static void pbs_free (PBSDATA *data)
{                               /* --- free statistics data */
  assert(data);                 /* check the function argument */
  if (data->sizes) free(data->sizes);
  if (data->bins)  free(data->bins);
  free(data->pp);               /* delete point process */
  free(data->heap);             /* and heap/support buffer */
}  /* pbs_free() */

/*--------------------------------------------------------------------*/

static int pbs_alloc (PBSDATA *data, TRAINSET *trnset, int stat,
                      double width, double beg, double end)
{                               /* --- allocate statistics data */
  ITEM   n;                     /* number of items */
  SUPP   l, r, c;               /* left and right bound of window */
  SUPP   m;                     /* maximum length of a train */
  size_t z;                     /* size of a buffer element */
  double e;                     /* end of test window */
  PPELEM *pp;                   /* trains as a point process */

  assert(data && trnset);       /* check the function arguments */
  data->trnset = trnset;        /* note the underlying train set */
  data->len    = end -beg;      /* and the length of the time range */
  data->pp = (PPELEM*)malloc((tns_extent(trnset)+1) *sizeof(PPELEM));
  if (!data->pp) return -1;     /* create a point process */
  z = 3*sizeof(SUPP);           /* determine the buffer element size */
  if (sizeof(HEAPELEM) > z) z = sizeof(HEAPELEM);
  n = tns_cnt(trnset);          /* create a heap for train merging */
  data->heap = (HEAPELEM*)malloc(((size_t)n+1) *z);
  if (!data->heap) { free(data->pp); return -1; }
  data->gfs = (SUPP*)data->heap;/* use it also as a buffer */
  data->cfs = data->gfs +n+1;   /* for the global and conditional */
  data->buf = data->cfs +n+1;   /* item frequencies/weights */
  data->bins  = NULL;           /* clear fields for easy clean-up */
  data->sizes = NULL;           /* (allows to call pbs_free()) */
  if (stat == STAT_CPO) {       /* if pattern overlap */
    tns2pp(trnset, data->pp, data->heap);
    l = r = c = 0; width *= 2;  /* init. window range indices */
    for (pp = data->pp; pp[l].pt < TRN_END; l++) { /* traverse points */
      e = pp[l].pt +width;      /* get end of current window */
      while (pp[r].pt <= e) r++;/* find first point after window */
      if (r-l > c) c = r-l;     /* update maximum number of points */
    }                           /* in a window of width 2*width */
    m = tns_maxlen(trnset)+1;   /* allocate point bin array */
    data->bins  = (ITEM**)malloc((size_t)m *sizeof(ITEM*));
    data->sizes = (ITEM*)malloc((size_t)(c+3) *(size_t)m *sizeof(ITEM));
    if (!data->bins || !data->sizes) { pbs_free(data); return -1; }
    data->bins[0] = data->sizes +m;
  }                             /* allocate sizes and contents arrays */
  return 0;                     /* return 'ok' */
}  /* pbs_alloc() */

/*--------------------------------------------------------------------*/

static void pbs_init (PBSDATA *data, ITEM item,
                      double width, int stat, int alpha)
{                               /* --- initialize statistics data */
  ITEM   i, n;                  /* loop variable, number of items */
  SUPP   l, r, m, k;            /* left and right bound of window */
  SUPP   size;                  /* number of distinct items in range */
  double b, e;                  /* beginning and end of point bin */
  PPELEM *pp;                   /* trains as a point process */
  TRAIN  *t;                    /* train to analyze */

  assert(data                   /* check the function arguments */
  &&    (item >= 0) && (item < tns_cnt(data->trnset))
  &&    (width > 0) && (alpha > 0));
  tns2pp(data->trnset, data->pp, data->heap);
  n  = tns_cnt(data->trnset);   /* clear the frequencies arrays */
  memset(data->gfs, 0, (size_t)(n+n+n+2) *sizeof(SUPP));
  pp = data->pp; l = r = 0;     /* get process, init. range indices */
  if ((stat == STAT_CPC) || (stat == STAT_CXC)
  ||  (stat == STAT_CCF) || (stat == STAT_CCR)) {
    for (m = 0; pp[m].pt < TRN_END; m++) {
      b = pp[m].pt -width;      /* traverse the points of the trains */
      e = pp[m].pt +width;      /* and get the point bin boundaries */
      while (pp[l].pt <  b) l++;/* find first point in    point bin */
      while (pp[r].pt <= e) r++;/* find first point after point bin */
      for (size = 0, k = l; k < r; k++)
        if (data->buf[pp[k].item]++ <= 0)
          size++;               /* count the items in the point bin */
      if (data->buf[item] > 0) size -= 1;
      data->gfs[size] += 1;     /* count the pattern complexity */
      for (k = l; k < r; k++)   /* clear the item flags/counters */
        data->buf[pp[k].item] = 0;
    }                           /* (faster than clearing all flags) */
    if (stat != STAT_CPC) {     /* if not cond. pattern complexity */
      for (data->avg = 0, i = 0; i < n; i++)
        data->avg += (double)i *(double)data->gfs[i]; }
    else {                      /* if conditional pattern complexity */
      for (data->avg = 0, i = 0; i < n; i++)
        data->avg += powi((double)i, alpha) *(double)data->gfs[i];
    }                           /* include exponent alpha */
    data->avg /= (double)m;     /* sum the pattern complexities */
  }                             /* and compute their average */
  if ((stat == STAT_CIF) || (stat == STAT_CIR)) {
    for (m = 0; pp[m].pt < TRN_END; m++) {
      b = pp[m].pt -width;      /* traverse the points of the trains */
      e = pp[m].pt +width;      /* and get the point bin boundaries */
      while (pp[l].pt <  b) l++;/* find first point in    point bin */
      while (pp[r].pt <= e) r++;/* find first point after point bin */
      for (size = 0, k = l; k < r; k++)
        data->buf[pp[k].item] = 1;
      data->buf[item] = 0;      /* count the items in the point bin */
      for (k = l; k < r; k++) { /* traverse the items in the bin */
        if (data->buf[i = pp[k].item] <= 0) continue;
        data->gfs[i] += 1;      /* if item has not yet been counted, */
        data->buf[i]  = 0;      /* count the item occurrence and */
      }                         /* clear the item flags/counters */
    }
  }
  if ((stat == STAT_CIW) || (stat == STAT_CWR)) {
    for (m = 0; pp[m].pt < TRN_END; m++) {
      b = pp[m].pt -width;      /* traverse the points of the trains */
      e = pp[m].pt +width;      /* and get the point bin boundaries */
      while (pp[l].pt <  b) l++;/* find first point in    point bin */
      while (pp[r].pt <= e) r++;/* find first point after point bin */
      for (size = 0, k = l; k < r; k++)
        if (data->buf[pp[k].item]++ <= 0)
          size++;               /* count the items in the point bin */
      if (data->buf[item] > 0){ /* check for test item, adapt size */
        data->buf[item] = 0; size -= 1; }
      for (k = l; k < r; k++) { /* traverse the items in the bin */
        if (data->buf[i = pp[k].item] <= 0) continue;
        data->gfs[i] += size;   /* if item has not yet been counted, */
        data->buf[i]  = 0;      /* sum the item weight and */
      }                         /* clear the item flags/counters */
    }
  }                             /* remove item from point process */
  t = tns_train(data->trnset, item);
  ppxtrn(data->pp, trn_item(t), NULL);
  data->cnt = (SUPP)tns_extent(data->trnset);
}  /* pbs_init() */             /* compute item probability */

/*--------------------------------------------------------------------*/

static void pbs_term (PBSDATA *data, ITEM item)
{ }                             /* --- terminate computations */

/*--------------------------------------------------------------------*/

static void pbs_cxfrq (PBSDATA *data, TRAIN *train, double width)
{                               /* --- collect complexity frequencies */
  ITEM   n;                     /* number of items */
  SUPP   l, r, k;               /* left and right bound of window */
  SUPP   size;                  /* number of distinct items in range */
  double b, e;                  /* beginning and end of point bin */
  PPELEM *pp;                   /* trains as a point process */
  const double *t;              /* to traverse points of the train */

  assert(data && train && (width >= 0)); /* check function arguments */
  n  = tns_cnt(data->trnset);   /* clear the complexity frequencies */
  memset(data->cfs, 0, (size_t)(n+n+1) *sizeof(SUPP));
  pp = data->pp; l = r = 0;     /* get process, init. range indices */
  for (t = trn_points(train); *t < TRN_END; t++) {
    b = *t -width;              /* traverse the points of the train */
    e = *t +width;              /* and get the point bin boundaries */
    while (pp[l].pt <  b) l++;  /* find first point in    point bin */
    if (r < l) r = l;           /* start at least with first point */
    while (pp[r].pt <= e) r++;  /* find first point after point bin */
    for (size = 0, k = l; k < r; k++)
      if (data->buf[pp[k].item]++ <= 0)
        size++;                 /* count the items in the point bin */
    data->cfs[size] += 1;       /* count the pattern complexity */
    for (k = l; k < r; k++)     /* clear the item flags/counters */
      data->buf[pp[k].item] = 0;/* (faster than clearing all flags) */
  }
  data->cond = trn_len(train);  /* note the number of point bins */
  data->prob = (double)data->cond /(double)data->cnt;
}  /* pbs_cxfrq() */

/*--------------------------------------------------------------------*/

static void pbs_itfrq (PBSDATA *data, TRAIN *train, double width)
{                               /* --- collect item frequencies */
  ITEM   i, n;                  /* item, number of items */
  SUPP   l, r, k;               /* left and right bound of window */
  double b, e;                  /* beginning and end of point bin */
  PPELEM *pp;                   /* trains as a point process */
  const double *t;              /* to traverse points of the train */

  assert(data && train && (width >= 0)); /* check function arguments */
  n  = tns_cnt(data->trnset);   /* clear the complexity frequencies */
  memset(data->cfs, 0, (size_t)(n+n+1) *sizeof(SUPP));
  pp = data->pp; l = r = 0;     /* get process, init. range indices */
  for (t = trn_points(train); *t < TRN_END; t++) {
    b = *t -width;              /* traverse the points of the train */
    e = *t +width;              /* and get the point bin boundaries */
    while (pp[l].pt <  b) l++;  /* find first point in    point bin */
    if (r < l) r = l;           /* start at least with first point */
    while (pp[r].pt <= e) r++;  /* find first point after point bin */
    for (k = l; k < r; k++)     /* traverse the points in the bin */
      data->buf[pp[k].item] = 1;/* and mark the occurring items */
    for (k = l; k < r; k++) {   /* traverse the items in the bin */
      if (data->buf[i = pp[k].item] <= 0) continue;
      data->cfs[i] += 1;        /* if item has not yet been counted, */
      data->buf[i]  = 0;        /* count the item occurrence and */
    }                           /* clear the item flags/counters */
  }
  data->cond = trn_len(train);  /* note the number of point bins */
  data->prob = (double)data->cond /(double)data->cnt;
}  /* pbs_itfrq() */

/*--------------------------------------------------------------------*/

static void pbs_itwgt (PBSDATA *data, TRAIN *train, double width)
{                               /* --- collect item weights */
  ITEM   i, n;                  /* item, number of items */
  SUPP   l, r, k;               /* left and right bound of window */
  SUPP   size;                  /* number of distinct items in range */
  double b, e;                  /* beginning and end of point bin */
  PPELEM *pp;                   /* trains as a point process */
  const double *t;              /* to traverse points of the train */

  assert(data && train && (width >= 0)); /* check function arguments */
  n  = tns_cnt(data->trnset);   /* clear the complexity frequencies */
  memset(data->cfs, 0, (size_t)(n+n+1) *sizeof(SUPP));
  pp = data->pp; l = r = 0;     /* get process, init. range indices */
  for (t = trn_points(train); *t < TRN_END; t++) {
    b = *t -width;              /* traverse the points of the train */
    e = *t +width;              /* and get the point bin boundaries */
    while (pp[l].pt <  b) l++;  /* find first point in    point bin */
    if (r < l) r = l;           /* start at least with first point */
    while (pp[r].pt <= e) r++;  /* find first point after point bin */
    for (size = 0, k = l; k < r; k++)
      if (data->buf[pp[k].item]++ <= 0)
        size++;                 /* count the items in the point bin */
    for (k = l; k < r; k++) {   /* traverse the items in the window */
      if (data->buf[i = pp[k].item] <= 0) continue;
      data->cfs[i] += size;     /* if item has not yet been counted, */
      data->buf[i]  = 0;        /* sum the item weight and */
    }                           /* clear the item flags/counters */
  }
  data->cond = trn_len(train);  /* note the number of point bins */
  data->prob = (double)data->cond /(double)data->cnt;
}  /* pbs_itwgt() */

/*--------------------------------------------------------------------*/

static double pbs_cpc (PBSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional pattern complexity */
  TID    n;                     /* number of points in train */
  SUPP   l, r, k;               /* left and right bound of window */
  SUPP   size;                  /* number of distinct items in range */
  double b, e;                  /* beginning and end of point bin */
  double sum;                   /* sum of pattern complexities */
  PPELEM *pp;                   /* trains as a point process */
  const double *t;              /* to traverse points of the train */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  memset(data->buf, 0, (size_t)tns_cnt(data->trnset) *sizeof(SUPP));
  sum = 0; l = r = 0;           /* clear statistic and range indices */
  pp  = data->pp;               /* get the point process */
  for (t = trn_points(train); *t < TRN_END; t++) {
    b = *t -width;              /* traverse the points of the train */
    e = *t +width;              /* and get the point bin boundaries */
    while (pp[l].pt <  b) l++;  /* find first point in    point bin */
    if (r < l) r = l;           /* start at least with first point */
    while (pp[r].pt <= e) r++;  /* find first point after point bin */
    for (size = 0, k = l; k < r; k++)
      if (data->buf[pp[k].item]++ <= 0)
        size++;                 /* count the items in the point bin */
    sum += powi((double)size, alpha);
    for (k = l; k < r; k++)     /* sum the pattern complexities */
      data->buf[pp[k].item] = 0;/* and clear the item flags */
  }
  n = trn_len(train);           /* get the number of points in train */
  if (n > 0) sum /= (double)n;  /* and compute the statistic value */
  return (data->avg > 0) ? (sum -data->avg) /data->avg : 0;
}  /* pbs_cpc() */

/*--------------------------------------------------------------------*/

static double pbs_cxc (PBSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional excess complexity */
  TID    n;                     /* number of points in train */
  SUPP   l, r, i;               /* left and right bound of window */
  SUPP   size;                  /* number of distinct items in range */
  double b, e;                  /* beginning and end of point bin */
  double sum, x;                /* sum of pattern complexities */
  PPELEM *pp;                   /* trains as a point process */
  const double *t;              /* to traverse points of the train */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  memset(data->buf, 0, (size_t)tns_cnt(data->trnset) *sizeof(SUPP));
  sum = 0; l = r = 0;           /* clear statistic and range indices */
  pp  = data->pp;               /* get the point process */
  for (t = trn_points(train); *t < TRN_END; t++) {
    b = *t -width;              /* traverse the points of the train */
    e = *t +width;              /* and get the point bin boundaries */
    while (pp[l].pt <  b) l++;  /* find first point in    point bin */
    if (r < l) r = l;           /* start at least with first point */
    while (pp[r].pt <= e) r++;  /* find first point after point bin */
    for (size = 0, i = l; i < r; i++)
      if (data->buf[pp[i].item]++ <= 0)
        size++;                 /* count the items in the point bin */
    x = (double)size -data->avg;
    if (x > 0) sum += powi(x, alpha);
    for (i = l; i < r; i++)     /* sum the excess complexities */
      data->buf[pp[i].item] = 0;/* and clear the item flags */
  }
  n = trn_len(train);           /* compute the statistic value */
  return (n > 0) ? sum /(double)n : 0;
}  /* pbs_cxc() */

/*--------------------------------------------------------------------*/

static double pbs_ccf (PBSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional complexity freq. */
  ITEM   i, n;                  /* loop variable, number of items */
  double sum, x;                /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  pbs_cxfrq(data, train, width);/* compute the complexity frequencies */
  n = tns_cnt(data->trnset);    /* traverse the pattern complexities */
  for (sum = 0, i = (int)ceil(data->avg); i <= n; i++) {
    x = (double)data->cfs[i] -data->prob *(double)data->gfs[i];
    if (x > 0) sum += x *powi((double)i -data->avg, alpha);
  }                             /* sum the excess frequencies */
  return (n > 1) ? sum /(double)n : 0;
}  /* pbs_ccf() */              /* finally normalize the statistic */

/*--------------------------------------------------------------------*/

static double pbs_ccr (PBSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- cond. complexity freq. ratio */
  ITEM   i, n;                  /* loop variable, number of items */
  double sum, x, y;             /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  pbs_cxfrq(data, train, width);/* compute the complexity frequencies */
  n = tns_cnt(data->trnset);    /* traverse the pattern complexities */
  for (sum = 0, i = (int)ceil(data->avg); i <= n; i++) {
    x = (double)data->cfs[i] -(y = data->prob *(double)data->gfs[i]);
    if (x > 0) sum += x/(y+1) *powi((double)i -data->avg, alpha);
  }                             /* sum the excess frequencies ratios */
  return (n > 1) ? sum /(double)n : 0;
}  /* pbs_ccr() */              /* finally normalize the statistic */

/*--------------------------------------------------------------------*/

static double pbs_cif (PBSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional item frequency */
  ITEM   i, n;                  /* loop variable, number of items */
  double sum, x;                /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  pbs_itfrq(data, train, width);/* compute item frequencies */
  n = tns_cnt(data->trnset);    /* and traverse the items */
  for (sum = 0, i = 0; i < n; i++) {
    #if 0
    x = (double)data->cfs[i] -data->prob *(double)data->gfs[i];
    #else
    x = (double)data->cfs[i] /(double)data->cond
      - (double)data->gfs[i] /(double)data->cnt;
    #endif
    if (x > 0) sum += powi(x, alpha);
  }                             /* sum the excess item frequencies */
  return (n > 1) ? sum /(double)n : 0;
}  /* pbs_cif() */              /* finally normalize the statistic */

/*--------------------------------------------------------------------*/

static double pbs_cir (PBSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional item freq. ratio */
  ITEM   i, n;                  /* loop variable, number of items */
  double sum, x, y;             /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  pbs_itfrq(data, train, width);/* compute item frequencies */
  n = tns_cnt(data->trnset);    /* and traverse the items */
  for (sum = 0, i = 0; i < n; i++) {
    #if 0
    x = (double)data->cfs[i] -(y = data->prob *(double)data->gfs[i]);
    #else
    y = (double)data->gfs[i] /(double)data->cnt;
    x = (double)data->cfs[i] /(double)data->cond -y;
    #endif
    if (x > 0) sum += powi(x/y, alpha);
  }                             /* sum excess item frequencies ratios */
  return (n > 1) ? sum /(double)n : 0;
}  /* pbs_cir() */              /* finally normalize the statistic */

/*--------------------------------------------------------------------*/

static double pbs_ciw (PBSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional item weight */
  ITEM   i, n;                  /* loop variable, number of items */
  double sum, x;                /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  pbs_itwgt(data, train, width);/* compute the item weights */
  n = tns_cnt(data->trnset);    /* and traverse the items */
  for (sum = 0, i = 0; i < n; i++) {
    #if 0
    x = (double)data->cfs[i] -data->prob *(double)data->gfs[i];
    #else
    x = (double)data->cfs[i] /(double)data->cond
      - (double)data->gfs[i] /(double)data->cnt;
    #endif
    if (x > 0) sum += powi(x, alpha);
  }                             /* sum the excess weights */
  return (n > 1) ? sum /(double)n : 0;
}  /* pbs_ciw() */              /* finally normalize the statistic */

/*--------------------------------------------------------------------*/

static double pbs_cwr (PBSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional item weight ratio */
  ITEM   i, n;                  /* loop variable, number of items */
  double sum, x, y;             /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  pbs_itwgt(data, train, width);/* compute the item weights */
  n = tns_cnt(data->trnset);    /* and traverse the items */
  for (sum = 0, i = 0; i < n; i++) {
    #if 0
    x = (double)data->cfs[i] -(y = data->prob *(double)data->gfs[i]);
    #else
    y = (double)data->gfs[i] /(double)data->cnt;
    x = (double)data->cfs[i] /(double)data->cond -y;
    #endif
    if (x > 0) sum += powi(x/y, alpha);
  }                             /* sum the excess weight ratios */
  return (n > 1) ? sum /(double)n : 0;
}  /* pbs_cwr() */              /* finally normalize the statistic */

/*--------------------------------------------------------------------*/

static double pbs_cpo (PBSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional pattern overlap */
  PPELEM *pp;                   /* trains as a point process */
  double b, e;                  /* beginning and end of point bin */
  TID    k, p, n;               /* loop variables, number of bins */
  SUPP   l, r, i;               /* left and right bound of range */
  ITEM   *s;                    /* to traverse items of a point bin */
  ITEM   m;                     /* number of items in a point bin */
  double sum;                   /* sum of pattern overlaps */
  const double *t;              /* to traverse points of the train */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  pp = data->pp; l = r = 0;     /* get process, clear range indices */
  s  = data->bins[n = 0];       /* get the first point bin */
  for (t = trn_points(train); *t < TRN_END; t++) {
    b = *t -width;              /* traverse the points of the train */
    e = *t +width;              /* and get the point bin boundaries */
    while (pp[l].pt <  b) l++;  /* find first point in    point bin */
    while (pp[r].pt <= e) r++;  /* find first point after point bin */
    for (i = l; i < r; i++)     /* collect the items */
      *s++ = pp[i].item;        /* in the point bin */
    *s++ = TA_END;              /* store a sentinel at the end */
    data->bins[++n] = s;        /* count the completed point bin */
  }                             /* and start the next point bin */
  if (n < 2) return 0;          /* check for at least two point bins */
  for (k = 0; k < n; k++) {     /* traverse the point bins */
    m = (ITEM)(data->bins[k+1] -data->bins[k] -1);
    ia_qsort(data->bins[k], (size_t)m, +1);
    data->sizes[k] = m = ia_unique(data->bins[k], (size_t)m);
    data->bins[k][m] = TA_END;  /* remove duplicate items from bins */
  }                             /* and store a sentinel at the end */
  memset(data->cfs, 0, (size_t)tns_cnt(data->trnset) *sizeof(SUPP));
  sum = 0;                      /* initialize sum and item flags */
  for (k = 1; k < n; k++) {     /* traverse the time bins */
    for (s = data->bins[k]; *s != TA_END; s++)
      data->cfs[*s] = 1;        /* set flags for contained items */
    for (p = k; --p >= 0; ) {   /* traverse preceding time bins */
      for (m = 0, s = data->bins[p]; *s != TA_END; s++)
        m += data->cfs[*s];     /* count items in time bin overlap */
      if (m > 1) sum += powi((double)m, alpha);
    }                           /* sum the pattern overlaps */
    for (s = data->bins[k]; *s != TA_END; s++)
      data->cfs[*s] = 0;        /* clear flags for contained items */
  }                             /* finally normalized summed overlaps */
  return 0.5*sum /((double)n*(double)(n-1));
}  /* pbs_cpo() */

/*--------------------------------------------------------------------*/

static int pbs_test (TRAINSET *trnset, ITEM item,
                     double width, double beg, double end,
                     int surr, RNG *rng, int stat, int alpha,
                     long runs, int qtls, FILE *out)
{                               /* --- execute statistical test */
  int       r = 0;              /* function result, error status */
  ITEM      i, n, e;            /* loop variables, buffer */
  long      k;                  /* loop variable for runs */
  TRAIN     *train, *t;         /* surrogate and original train */
  size_t    z, points;          /* (total) number of points/spikes */
  TRNSURRFN *surrfn;            /* surrogate train generation func. */
  PBSTATFN  *pbsfn;             /* test statistic to compute */
  double    *stats = NULL;      /* train statistic values */
  double    *sums  = NULL;      /* sums of statistic values */
  double    x, y, total;        /* (total value of) test statistic */
  double    p;                  /* computed p-value */
  CCHAR     *s;                 /* name/identifier of an item/neuron */
  PBSDATA   data;               /* point bins statistics data */

  assert(trnset && (width >= 0) && (end >= beg) && rng
  &&    (alpha > 0) && out);    /* check the function arguments */

  /* --- initialize --- */
  train = trn_create(0, NULL, tns_maxlen(trnset));
  if (!train) return -1;        /* create a train for surrogates */
  if (runs > 0) {               /* if to generate surrogates */
    stats = (double*)malloc((size_t)(runs +runs) *sizeof(double));
    if (!stats) { trn_delete(train); return -1; }
    sums = stats +runs;         /* allocate the statistics arrays */
    dbl_clear(sums, (size_t)runs);
  }                             /* clear the statistic sums */
  total = 0; points = 0;        /* initialize the statistic total */
  if (pbs_alloc(&data, trnset, stat, width, beg, end) != 0) {
    if (stats) { free(stats); } trn_delete(train); return -1; }
  if      (surr == 'r')      surrfn = trn_random;
  else if (surr == 'd')      surrfn = trn_dither;
  else if (surr == 's')      surrfn = trn_shift;
  else                       surrfn = trn_random;
  if      (stat == STAT_CPC) pbsfn  = pbs_cpc;
  else if (stat == STAT_CXC) pbsfn  = pbs_cxc;
  else if (stat == STAT_CCF) pbsfn  = pbs_ccf;
  else if (stat == STAT_CCR) pbsfn  = pbs_ccr;
  else if (stat == STAT_CIF) pbsfn  = pbs_cif;
  else if (stat == STAT_CIR) pbsfn  = pbs_cir;
  else if (stat == STAT_CIW) pbsfn  = pbs_ciw;
  else if (stat == STAT_CWR) pbsfn  = pbs_cwr;
  else if (stat == STAT_CPO) pbsfn  = pbs_cpo;
  else                       pbsfn  = pbs_cpc;
  header(out, runs, qtls, " %11s");

  /* --- execute tests for items --- */
  n = tns_cnt(trnset);          /* get the number of trains */
  if (item >= 0) e = (i = item)+1;
  else         { e = n; i = 0; }/* get the range of items */
  for ( ; i < e; i++) {         /* traverse the items to analyze */
    t = tns_train(trnset, i);   /* sum the number of points */
    points += z = (size_t)trn_len(t);
    pbs_init(&data, i, width, stat, alpha);
    x = (z > 0) ? pbsfn(&data, t, width, alpha) : 0;
    total += x;                 /* compute and sum the test statistic */
    s = ib_name(tns_base(trnset), i);
    if (runs <= 0) {            /* if no surrogate data requested, */
      fprintf(out, "%6s  %11.6f", s, x); /* report statistic value */
      if (qtls) fprintf(out, " %6"SIZE_FMT, z); }
    else {                      /* if to compare to surrogate data */
      for (p = 0, k = runs; --k >= 0; ) {
        train = surrfn(tns_train(trnset, i), beg, end, rng, train);
        trn_sort(train, +1);    /* create a surrogate train */
        y = (z > 0) ? pbsfn(&data, train, width, alpha) : 0;
        stats[k] = y;           /* compute and store test statistic */
        sums[k] += y;           /* sum test statistic over items */
        if (y >= x) p++;        /* count the values exceeding or */
      }                         /* meeting the actual statistic value */
      p /= (double)runs;        /* compute and print the p-value */
      fprintf(out, "%6s  %8.6f%c", s, p, sigind(p));
      if (qtls) {               /* if extended output requested */
        dbl_qsort(stats, (size_t)runs, +1);
        report(out, " %11.6f", stats, runs, x, z);
      }                         /* sort statistic values and */
    }                           /* report quantiles of statistic */
    fprintf(out, "\n");         /* and the actual statistic value, */
    pbs_term(&data, i);         /* then terminate the output line */
  }                             /* and the statistics computation */

  /* --- total statistics --- */
  if ((runs > 0) && (item < 0)) /* if all neurons were checked */
    footer(out, sums, runs, total, points, qtls);

  /* --- clean up --- */
  pbs_free(&data);              /* delete the work memory, */
  if (stats) free(stats);       /* the statistic arrays */
  trn_delete(train);            /* and the surrogate train */
  return r;                     /* return the error status */
}  /* pbs_test() */

/*----------------------------------------------------------------------
  Covering Window Test Statistics
----------------------------------------------------------------------*/

static void cws_free (CWSDATA *data)
{                               /* --- free statistics data */
  assert(data);                 /* check the function argument */
  free(data->pp);               /* delete point process */
  free(data->heap);             /* and heap/support buffer */
}  /* cws_free() */

/*--------------------------------------------------------------------*/

static int cws_alloc (CWSDATA *data, TRAINSET *trnset, int stat)
{                               /* --- allocate statistics data */
  ITEM   n;                     /* number of items */
  size_t z;                     /* size of a buffer element */

  assert(data && trnset);       /* check the function arguments */
  data->trnset = trnset;        /* note the underlying train set */
  data->pp = (PPELEM*)malloc((tns_extent(trnset)+1) *sizeof(PPELEM));
  if (!data->pp) return -1;     /* create a point process */
  z = 3*sizeof(SUPP);           /* determine the buffer element size */
  if (sizeof(HEAPELEM) > z) z = sizeof(HEAPELEM);
  n = tns_cnt(trnset);          /* create a heap for train merging */
  data->heap = (HEAPELEM*)malloc(((size_t)n+1) *z);
  if (!data->heap) { free(data->pp); return -1; }
  data->gfs = (SUPP*)data->heap;/* use it also as a buffer */
  data->cfs = data->gfs +n+1;   /* for the global and conditional */
  data->buf = data->cfs +n+1;   /* item frequencies/weights */
  return 0;                     /* return 'ok' */
}  /* cws_alloc() */

/*--------------------------------------------------------------------*/

static void cws_init (CWSDATA *data, ITEM item,
                      double width, int stat, int alpha)
{                               /* --- initialize statistics data */
  assert(data                   /* check the function arguments */
  &&    (item >= 0) && (item < tns_cnt(data->trnset))
  &&    (width > 0) && (alpha > 0));
  tns2pp(data->trnset, data->pp, data->heap);
}  /* cws_init() */

/*--------------------------------------------------------------------*/

static void cws_term (CWSDATA *data, ITEM item)
{ }                             /* --- terminate computations */

/*--------------------------------------------------------------------*/

static void cws_cxfrq (CWSDATA *data, TRAIN *train, double width)
{                               /* --- collect complexity frequencies */
  ITEM   item, n;               /* item, number of items/trains */
  SUPP   l, r;                  /* left and right bound of window */
  SUPP   c, g;                  /* global and cond. window counter */
  SUPP   size;                  /* number of distinct items in range */
  double end;                   /* end of current window */
  SUPP   *gxs, *cxs, *buf;      /* complexity distributions */
  PPELEM *pp;                   /* trains as a point process */

  assert(data && train && (width >= 0)); /* check function arguments */
  ppxtrn(data->pp, -1, train);  /* update point process with train */
  n    = tns_cnt(data->trnset); /* get the number of trains */
  gxs  = (SUPP*)memset(data->gfs, 0, (size_t)(n+n+n+2) *sizeof(SUPP));
  cxs  = data->cfs;             /* clear global and conditional */
  buf  = data->buf;             /* complexity frequency arrays */
  item = trn_item(train);       /* get the item to test */
  size = l = r = g = c = 0;     /* clear size and range indices */
  for (pp = data->pp; pp[l].pt < TRN_END; l++) { /* traverse points */
    end = pp[l].pt +width;      /* get end of current window */
#if MAXWIN
    if (pp[r].pt <= end) {      /* if the window is maximal */
#endif
    while (pp[r].pt <= end)     /* while still inside current window */
      if (buf[pp[r++].item]++ <= 0)
        size++;                 /* add next point to the window */
    if (buf[item] > 0) { gxs[size-1]++; cxs[size-1]++; c += 1; }
    else               { gxs[size]++; }
    g += 1;                     /* update complexity counters */
#if MAXWIN
    }
#endif
    if (--buf[pp[l].item] <= 0) /* remove the first point */
      size--;                   /* from the current window */
  }                             /* and adapt the set size */
  data->cnt  = g;               /* note numbers of windows and */
  data->cond = c;               /* compute occurrence probability */
  data->prob = (double)c/(double)g;
}  /* cws_cxfrq() */

/*--------------------------------------------------------------------*/

static void cws_itfrq (CWSDATA *data, TRAIN *train, double width)
{                               /* --- collect item frequencies */
  ITEM   item, i, n;            /* item, number of items/trains */
  SUPP   l, r, k;               /* left and right bound of window */
  SUPP   c, g;                  /* global and cond. window counter */
  SUPP   y, z;                  /* conditional increment, buffer */
  double end;                   /* end of current window */
  SUPP   *gfs, *cfs, *buf;      /* point frequency distributions */
  PPELEM *pp;                   /* trains as a point process */

  assert(data && train && (width >= 0)); /* check function arguments */
  ppxtrn(data->pp, -1, train);  /* update point process with train */
  n    = tns_cnt(data->trnset); /* get the number of trains */
  gfs  = (SUPP*)memset(data->gfs, 0, (size_t)(n+n+n+2) *sizeof(SUPP));
  cfs  = data->cfs;             /* clear global and conditional */
  buf  = data->buf;             /* item frequency arrays */
  item = trn_item(train);       /* get the item to test */
  l = r = g = c = 0;            /* clear the range indices */
  for (pp = data->pp; pp[l].pt < TRN_END; l++) { /* traverse points */
    end = pp[l].pt +width;      /* get end of current window */
#if MAXWIN
    if (pp[r].pt <= end) {      /* if the window is maximal */
#endif
    while (pp[r].pt <= end)     /* while still inside current window */
      buf[pp[r++].item]++;      /* add next point to the window */
    if (buf[item] > 0) c += y = 1;
    else y = 0;                 /* get the conditional increment */
    g += 1;                     /* count the window globally */
    for (k = l; k < r; k++) {   /* traverse the items in the window */
      z = buf[i = pp[k].item];  /* get the item occurrence counter */
      if (z < 0) continue;      /* if item is already counted, skip */
      buf[i] = -z;              /* mark item as counted */
      gfs[i] += 1;              /* always count an item globally */
      cfs[i] += y;              /* count item conditionally only */
    }                           /* if the test item is in the window */
    for (k = l; k < r; k++) {   /* unmark all items in the window */
      z = buf[i = pp[k].item]; if (z < 0) buf[i] = -z; }
#if MAXWIN
    }
#endif
    buf[pp[l].item]--;          /* remove the first point */
  }                             /* from the current window */
  gfs[item]  = cfs[item] = 0;   /* test item is not considered */
  data->cnt  = g;               /* note numbers of windows and */
  data->cond = c;               /* compute occurrence probability */
  data->prob = (double)c/(double)g;
}  /* cws_itfrq() */

/*--------------------------------------------------------------------*/

static void cws_itwgt (CWSDATA *data, TRAIN *train, double width)
{                               /* --- collect point weights */
  ITEM   item, i, n;            /* item, number of items/trains */
  SUPP   l, r, k;               /* left and right bound of window */
  SUPP   c, g;                  /* global and cond. window counter */
  SUPP   size;                  /* number of distinct items in range */
  SUPP   x, y, z;               /* global and conditional increment */
  double end;                   /* end of current window */
  SUPP   *gws, *cws, *buf;      /* point weight distributions */
  PPELEM *pp;                   /* trains as a point process */

  assert(data && train && (width >= 0)); /* check function arguments */
  ppxtrn(data->pp, -1, train);  /* update point process with train */
  n    = tns_cnt(data->trnset); /* get the number of trains */
  gws  = (SUPP*)memset(data->gfs, 0, (size_t)(n+n+n+2) *sizeof(SUPP));
  cws  = data->cfs;             /* clear global and conditional */
  buf  = data->buf;             /* item frequency arrays */
  item = trn_item(train);       /* get the item to test */
  size = l = r = g = c = 0;     /* clear size and range indices */
  for (pp = data->pp; pp[l].pt < TRN_END; l++) { /* traverse points */
    end = pp[l].pt +width;      /* get end of current window */
#if MAXWIN
    if (pp[r].pt <= end) {      /* if the window is maximal */
#endif
    while (pp[r].pt <= end)     /* while still inside current window */
      if (buf[pp[r++].item]++ <= 0)
        size++;                 /* add next point to the window */
    if (buf[item] > 0) { x = y = size-1; c += 1; }
    else               { x = size; y = 0; }
    g += 1;                     /* get the conditional increments */
    for (k = l; k < r; k++) {   /* traverse the items in the window */
      z = buf[i = pp[k].item];  /* get the item occurrence counter */
      if (z < 0) continue;      /* if item is already counted, skip */
      buf[i] = -z;              /* mark the item as counted */
      gws[i] += x;              /* always count an item globally */
      cws[i] += y;              /* count item conditionally only if */
    }                           /* the test item is in the window */
    for (k = l; k < r; k++) {   /* unmark all items in the window */
      z = buf[i = pp[k].item]; if (z < 0) buf[i] = -z; }
#if MAXWIN
    }
#endif
    if (--buf[pp[l].item] <= 0) /* remove the first point */
      size--;                   /* from the current window */
  }                             /* and adapt the set size */
  gws[item]  = cws[item] = 0;   /* test item is not considered */
  data->cnt  = g;               /* note numbers of windows and */
  data->cond = c;               /* compute occurrence frequency */
  data->prob = (double)c/(double)g;
}  /* cws_itwgt() */

/*--------------------------------------------------------------------*/

static double cws_cpc (CWSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional pattern complexity */
  ITEM   i, n;                  /* item, number of items/trains */
  SUPP   l, r;                  /* left and right bound of window */
  SUPP   size;                  /* number of distinct items in range */
  SUPP   g, c;                  /* global and cond. window counter */
  double end;                   /* end of current window */
  double gpc, cpc, x;           /* pattern complexities */
  PPELEM *pp;                   /* trains as a point process */
  SUPP   *buf;                  /* buffer for support values */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  ppxtrn(data->pp, -1, train);  /* update point process with train */
  n    = tns_cnt(data->trnset); /* get the number of trains */
  i    = trn_item(train);       /* and the item of the train */
  buf  = (SUPP*)memset(data->buf, 0, (size_t)n *sizeof(SUPP));
  gpc  = cpc = 0;               /* init. pattern complexities */
  size = l = r = g = c = 0;     /* clear size and range indices */
  for (pp = data->pp; pp[l].pt < TRN_END; l++) { /* traverse points */
    end = pp[l].pt +width;      /* get end of current window */
#if MAXWIN
    if (pp[r].pt <= end) {      /* if the window is maximal */
#endif
    while (pp[r].pt <= end)     /* while still inside window */
      if (buf[pp[r++].item]++ <= 0)
        size++;                 /* add next point to the window */
    if (buf[i] > 0) { cpc += x = powi((double)size-1, alpha); c += 1; }
    else            {        x = powi((double)size,   alpha); }
    g   += 1;                   /* count the window globally */
    gpc += x;                   /* sum the pattern complexities */
#if MAXWIN
    }
#endif
    if (--buf[pp[l].item] <= 0) /* remove the first point */
      size--;                   /* from the current window */
  }                             /* and adapt the set size */
  gpc /= (double)g;             /* compute average pattern complexity */
  cpc /= (double)c;             /* (globally and conditionally) */
  return (gpc > 0) ? (cpc-gpc) /gpc : 0;
}  /* cws_cpc() */

/*--------------------------------------------------------------------*/

static double cws_cxc (CWSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional excess complexity */
  ITEM   i, n;                  /* item, number of items/trains */
  SUPP   l, r;                  /* left and right bound of window */
  SUPP   size;                  /* number of distinct items in range */
  SUPP   c;                     /* global and cond. window counter */
  double end;                   /* end of current window */
  double gpc, sum, x;           /* pattern complexities */
  PPELEM *pp;                   /* trains as a point process */
  SUPP   *buf;                  /* buffer for support values */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  ppxtrn(data->pp, -1, train);  /* update point process with train */
  n    = tns_cnt(data->trnset); /* get the number of trains */
  i    = trn_item(train);       /* and the item of the train */
  buf  = (SUPP*)memset(data->buf, 0, (size_t)n *sizeof(SUPP));
  sum  = 0;                     /* init. pattern complexity */
  size = l = r = c = 0;         /* clear size and range indices */
  for (pp = data->pp; pp[l].pt < TRN_END; l++) { /* traverse points */
    end = pp[l].pt +width;      /* get end of current window */
#if MAXWIN
    if (pp[r].pt <= end) {      /* if the window is maximal */
#endif
    while (pp[r].pt <= end)     /* while not at end of window */
      if (buf[pp[r++].item]++ <= 0)
        size++;                 /* add next point to the window */
    sum += (double)((buf[i] > 0) ? size-1 : size);
    c   += 1;                   /* sum the pattern complexities */
#if MAXWIN
    }
#endif
    if (--buf[pp[l].item] <= 0) /* remove the first point */
      size--;                   /* from the current window */
  }                             /* and adapt the set size */
  gpc  = sum/(double)c;         /* compute global pattern complexity */
  sum  = 0;                     /* init. excess complexity */
  size = l = r = c = 0;         /* clear size and range indices */
  for ( ; pp[l].pt < TRN_END; l++) { /* traverse the points */
    end = pp[l].pt +width;      /* get end of current window */
#if MAXWIN
    if (pp[r].pt <= end) {      /* if the window is maximal */
#endif
    while (pp[r].pt <= end)     /* while not at end of window */
      if (buf[pp[r++].item]++ <= 0)
        size++;                 /* add next point to the window */
    if (buf[i] > 0) {           /* if the test item is contained */
      c += 1;                   /* count the window and */
      x = (double)(size-1)-gpc; /* compute the excess complexity */
      if (x > 0) sum += powi(x, alpha);
    }                           /* sum the statistic term if positive */
#if MAXWIN
    }
#endif
    if (--buf[pp[l].item] <= 0) /* remove the first point */
      size--;                   /* from the current window */
  }                             /* and adapt the set size */
  return (c > 0) ? sum/(double)c : 0;
}  /* cws_cxc() */              /* compute excess complexity */

/*--------------------------------------------------------------------*/

static double cws_ccf (CWSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional complexity freq. */
  ITEM   i, n;                  /* loop variable, number of items */
  double gpc;                   /* average global pattern complexity */
  double sum, x;                /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  cws_cxfrq(data, train, width);/* compute the complexity frequencies */
  n = tns_cnt(data->trnset);    /* traverse the pattern complexities */
  for (gpc = 0, i = 1; i <= n; i++)
    gpc += (double)i *(double)data->gfs[i];
  gpc /= (double)data->cnt;     /* compute average pattern complexity */
  for (sum = 0, i = (int)ceil(gpc); i <= n; i++) {
    x = (double)data->cfs[i] -data->prob *(double)data->gfs[i];
    if (x > 0) sum += x *powi((double)i -gpc, alpha);
  }                             /* sum the excess frequencies */
  return (n > 1) ? sum /(double)n : 0;
}  /* cws_ccf() */

/*--------------------------------------------------------------------*/

static double cws_ccr (CWSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- cond. complexity freq. ratio */
  ITEM   i, n;                  /* loop variable, number of items */
  double gpc;                   /* average global pattern complexity */
  double sum, x, y;             /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  cws_cxfrq(data, train, width);/* compute the complexity frequencies */
  n = tns_cnt(data->trnset);    /* traverse the pattern complexities */
  for (gpc = 0, i = 1; i <= n; i++)
    gpc += (double)i *(double)data->gfs[i];
  gpc /= (double)data->cnt;     /* compute average pattern complexity */
  for (sum = 0, i = (int)ceil(gpc); i <= n; i++) {
    x = (double)data->cfs[i] -(y = data->prob *(double)data->gfs[i]);
    if (x > 0) sum += x/(y+1) *powi((double)i -gpc, alpha);
  }                             /* sum the excess frequencies */
  return (n > 1) ? sum /(double)n : 0;
}  /* cws_ccr() */

/*--------------------------------------------------------------------*/

static double cws_cif (CWSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional item frequency */
  ITEM   i, n;                  /* loop variable, number of items */
  double sum, x;                /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  cws_itfrq(data, train, width);/* compute the item frequencies */
  n = tns_cnt(data->trnset);    /* and traverse the items */
  for (sum = 0, i = 0; i < n; i++) {
    #if 0
    x = (double)data->cfs[i] -data->prob *(double)data->gfs[i];
    #else
    x = (double)data->cfs[i] /(double)data->cond
      - (double)data->gfs[i] /(double)data->cnt;
    #endif
    if (x > 0) sum += powi(x, alpha);
  }                             /* sum the excess frequencies */
  return (n > 1) ? sum /(double)n : 0;
}  /* cws_cif() */

/*--------------------------------------------------------------------*/

static double cws_cir (CWSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional item freq. ratio */
  ITEM   i, n;                  /* loop variable, number of items */
  double sum, x, y;             /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  cws_itfrq(data, train, width);/* compute the item frequencies */
  n = tns_cnt(data->trnset);    /* and traverse the items */
  for (sum = 0, i = 0; i < n; i++) {
    #if 0
    x = (double)data->cfs[i] -(y = data->prob *(double)data->gfs[i]);
    #else
    y = (double)data->gfs[i] /(double)data->cnt;
    x = (double)data->cfs[i] /(double)data->cond -y;
    #endif
    if (x > 0) sum += powi(x/y, alpha);
  }                             /* sum the excess frequencies */
  return (n > 1) ? sum /(double)n : 0;
}  /* cws_cir() */

/*--------------------------------------------------------------------*/

static double cws_ciw (CWSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional item weight */
  ITEM   i, n;                  /* loop variable, number of items */
  double sum, x;                /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  cws_itwgt(data, train, width);/* compute the item weights */
  n = tns_cnt(data->trnset);    /* and traverse the items */
  for (sum = 0, i = 0; i < n; i++) {
    #if 0
    x = (double)data->cfs[i] -data->prob *(double)data->gfs[i];
    #else
    x = (double)data->cfs[i] /(double)data->cond
      - (double)data->gfs[i] /(double)data->cnt;
    #endif
    if (x > 0) sum += powi(x, alpha);
  }                             /* sum the excess frequencies */
  return (n > 1) ? sum /(double)n : 0;
}  /* cws_ciw() */

/*--------------------------------------------------------------------*/

static double cws_cwr (CWSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional item weight ratio */
  ITEM   i, n;                  /* loop variable, number of items */
  double sum, x, y;             /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  cws_itwgt(data, train, width);/* compute the item weights */
  n = tns_cnt(data->trnset);    /* and traverse the items */
  for (sum = 0, i = 0; i < n; i++) {
    #if 0
    x = (double)data->cfs[i] -(y = data->prob *(double)data->gfs[i]);
    #else
    y = (double)data->gfs[i] /(double)data->cnt;
    x = (double)data->cfs[i] /(double)data->cond -y;
    #endif
    if (x > 0) sum += powi(x/y, alpha);
  }                             /* sum the excess frequencies */
  return (n > 1) ? sum /(double)n : 0;
}  /* cws_cwr() */

/*--------------------------------------------------------------------*/

static int cws_test (TRAINSET *trnset, ITEM item,
                     double width, double beg, double end,
                     int surr, RNG *rng, int stat, int alpha,
                     long runs, int qtls, FILE *out)
{                               /* --- execute statistical test */
  int       r = 0;              /* function result, error status */
  ITEM      i, n, e;            /* loop variables, buffer */
  long      k;                  /* loop variable for runs */
  TRAIN     *train, *t;         /* surrogate and original train */
  size_t    z, points;          /* (total) number of points/spikes */
  TRNSURRFN *surrfn;            /* surrogate train generation func. */
  CWSTATFN  *cwsfn;             /* test statistic to compute */
  double    *stats = NULL;      /* train statistic values */
  double    *sums  = NULL;      /* sums of statistic values */
  double    x, y, total;        /* (total value of) test statistic */
  double    p;                  /* computed p-value */
  CCHAR     *s;                 /* name/identifier of an item/neuron */
  CWSDATA   data;               /* window statistics data */

  assert(trnset && (width >= 0) && (end >= beg) && rng
  &&    (alpha > 0) && out);    /* check the function arguments */

  /* --- initialize --- */
  train = trn_create(0, NULL, tns_maxlen(trnset));
  if (!train) return -1;        /* create a train for surrogates */
  if (runs > 0) {               /* if to generate surrogates */
    stats = (double*)malloc((size_t)(runs +runs) *sizeof(double));
    if (!stats) { trn_delete(train); return -1; }
    sums = stats +runs;         /* allocate the statistics arrays */
    dbl_clear(sums, (size_t)runs);
  }                             /* clear the statistic sums */
  total = 0; points = 0;        /* initialize the statistic total */
  if (cws_alloc(&data, trnset, stat) != 0) {
    if (stats) { free(stats); } trn_delete(train); return -1; }
  if      (surr == 'r')      surrfn = trn_random;
  else if (surr == 'd')      surrfn = trn_dither;
  else if (surr == 's')      surrfn = trn_shift;
  else                       surrfn = trn_random;
  if      (stat == STAT_CPC) cwsfn  = cws_cpc;
  else if (stat == STAT_CXC) cwsfn  = cws_cxc;
  else if (stat == STAT_CCF) cwsfn  = cws_ccf;
  else if (stat == STAT_CCR) cwsfn  = cws_ccr;
  else if (stat == STAT_CIF) cwsfn  = cws_cif;
  else if (stat == STAT_CIR) cwsfn  = cws_cir;
  else if (stat == STAT_CIW) cwsfn  = cws_ciw;
  else if (stat == STAT_CWR) cwsfn  = cws_cwr;
  else                       cwsfn  = cws_cpc;
  header(out, runs, qtls, " %11s");

  /* --- execute tests for items --- */
  n = tns_cnt(trnset);          /* get the number of trains */
  if (item >= 0) e = (i = item)+1;
  else         { e = n; i = 0; }/* get the range of items */
  for ( ; i < e; i++) {         /* traverse the items to analyze */
    t = tns_train(trnset, i);   /* sum the number of points */
    points += z = (size_t)trn_len(t);
    cws_init(&data, i, width, stat, alpha);
    x = (z > 0) ? cwsfn(&data, t, width, alpha) : 0;
    total += x;                 /* compute and sum the test statistic */
    s = ib_name(tns_base(trnset), i);
    if (runs <= 0) {            /* if no surrogate data requested, */
      fprintf(out, "%6s  %11.6f", s, x); /* report statistic value */
      if (qtls) fprintf(out, " %6"SIZE_FMT, z); }
    else {                      /* if to compare to surrogate data */
      for (p = 0, k = runs; --k >= 0; ) {
        train = surrfn(tns_train(trnset, i), beg, end, rng, train);
        trn_sort(train, +1);    /* create a surrogate train */
        y = (z > 0) ? cwsfn(&data, train, width, alpha) : 0;
        stats[k] = y;           /* compute and store test statistic */
        sums[k] += y;           /* sum test statistic over items */
        if (y >= x) p++;        /* count the values exceeding or */
      }                         /* meeting the actual statistic value */
      p /= (double)runs;        /* compute and print the p-value */
      fprintf(out, "%6s  %8.6f%c", s, p, sigind(p));
      if (qtls) {               /* if extended output requested */
        dbl_qsort(stats, (size_t)runs, +1);
        report(out, " %11.6f", stats, runs, x, z);
      }                         /* sort statistic values and */
    }                           /* report quantiles of statistic */
    fprintf(out, "\n");         /* and the actual statistic value, */
    cws_term(&data, i);         /* then terminate the output line */
  }                             /* and the statistic computation */

  /* --- total statistics --- */
  if ((runs > 0) && (item < 0)) /* if all neurons were checked */
    footer(out, sums, runs, total, points, qtls);

  /* --- clean up --- */
  cws_free(&data);              /* delete the work memory, */
  if (stats) free(stats);       /* the statistic arrays */
  trn_delete(train);            /* and the surrogate train */
  return r;                     /* return the error status */
}  /* cws_test() */

/*----------------------------------------------------------------------
  Influence Region Statistics
----------------------------------------------------------------------*/

static double ivs7trn (IVAL *dst, IVAL *src, TRAIN *train, double width)
{                               /* -- intersect intervals and train */
  double r = 0;                 /* total length of intervals */
  double b, e;                  /* beginning and end of intersection */
  IVAL   *d, *c;                /* to traverse the intervals */
  const double *t;              /* to traverse the train */

  assert(dst                    /* check the function arguments */
  &&     src && train && (width >= 0));
  t = trn_points(train);        /* get the points of the train */
  if (!(src->beg < INFINITY) || (*t >= TRN_END)) {
    dst->beg = dst->end = INFINITY;
    return 0;                   /* if intervals or train are empty, */
  }                             /* store a sentinel and abort */
  width = 0.5*width;            /* get +/- offset from region width */
  b = *t -width;                /* get beginning and end of */
  e = *t +width;                /* first point's influence region */
  c = d = dst;                  /* initialize the destination and */
  c->end = -INFINITY;           /* set the initial interval end */
  while (1) {                   /* interval list filtering loop */
    if (b > src->end) {         /* if the intervals do not overlap */
      if (!((++src)->beg < INFINITY))
        break; }                /* skip the processed interval */
    else {                      /* if the intervals may overlap */
      if (e >= src->beg) {      /* if the intervals overlap */
        if (src->beg > b) b = src->beg;
        if (src->end < e) e = src->end;
        if (b <= c->end) { c->end = e; }
        else { if (c < d) r += c->end -c->beg;
               d->beg = b; d->end = e; c = d++; }
      }                         /* expand old or store new interval */
      if (*++t >= TRN_END) break;
      b = *t -width;            /* skip the processed point, */
      e = *t +width;            /* check for the end of the train, */
    }                           /* and get beginning and end of */
  }                             /* next point's influence region */
  if (c < d) r += c->end -c->beg;
  d->beg = d->end = INFINITY;   /* store a sentinel interval */
  return r;                     /* return the number of used points */
}  /* ivs7trn() */

/*--------------------------------------------------------------------*/

static int stepcmp (const void *s1, const void *s2)
{                               /* --- compare two function steps */
  if (((FNSTEP*)s1)->pt < ((FNSTEP*)s2)->pt) return -1;
  if (((FNSTEP*)s1)->pt > ((FNSTEP*)s2)->pt) return +1;
  return 0;                     /* return diff. of points/spike times */
}  /* stepcmp() */

/*--------------------------------------------------------------------*/

static size_t stepfn (FNSTEP *steps, TRAINSET *trnset,
                      ITEM item, double width)
{                               /* --- build the step function */
  FNSTEP *s, *d;                /* to traverse the steps */
  double b, e;                  /* beginning and end of interval */
  diff_t h;                     /* height of the current step */
  ITEM   i;                     /* loop variable for items/neurons */
  const double *t;              /* to traverse the points/spikes */

  assert(steps                  /* check the function arguments */
  &&     trnset && (width >= 0));
  width *= 0.5;                 /* compute offset from region width */
  s = steps;                    /* get the destination steps and */
  s->pt = -INFINITY; s->ht = 1; /* store a sentinel at the start */
  for (i = tns_cnt(trnset); --i >= 0; ) {
    if (i == item) continue;    /* traverse items other than given */
    t = trn_points(tns_train(trnset, i));
    if (*t >= TRN_END) continue;/* skip points before start/after end */
    (++s)->pt = *t -width; s->ht = +1;
    (++s)->pt = *t +width; s->ht = -1; /* store the first interval */
    while (*++t < TRN_END) {    /* traverse the points */
      b = *t -width;            /* get beginning and end of */
      e = *t +width;            /* next point's influence region */
      if (b <= s->pt)           /* if overlap with previous interval, */
        s->pt = e;              /* merely update the right border */
      else {                    /* if no overlap with prev. interval */
        (++s)->pt = b; s->ht = +1;
        (++s)->pt = e; s->ht = -1;
      }                         /* store the new interval */
    }                           /* collect intervals around points */
  }                             /* over all trains other than item */
  (++s)->pt = INFINITY;         /* store a sentinel at the end */
  s->ht     = 0;                /* of the interval list */
  qsort(steps+1, (size_t)(s-steps)-1, sizeof(FNSTEP), stepcmp);
  for (d = steps, s = d+1; s->pt < INFINITY; s++) {
    if      (d->pt == s->pt) d->ht += s->ht;
    else if (d->ht != 0)     *++d = *s;
    else                     *d   = *s;
  }                             /* combine steps at the same point */
  (++d)->pt = INFINITY;         /* store a sentinel at the end */
  steps->ht = d->ht = 0;        /* of the interval list */
  for (h = 0, s = steps+1; s->pt < INFINITY; s++)
    s->ht = h += s->ht;         /* compute the step heights */
  assert(h <= 0);               /* check function for consistency */
  return (size_t)(s-steps);     /* return the created step function */
}  /* stepfn() */

/*--------------------------------------------------------------------*/

static void powsteps (FNSTEP *steps, int alpha)
{                               /* --- raise steps to given power */
  FNSTEP *s;                    /* to traverse the steps */

  assert(steps && (alpha > 0)); /* check the function arguments */
  if (alpha <= 1) return;       /* check for a power > 1 */
  for (s = steps+1; s->pt < INFINITY; s++)
    s->ht = ipowi(s->ht,alpha); /* raise all steps to given power */
}  /* powsteps() */

/*--------------------------------------------------------------------*/

static double austeps (FNSTEP *steps, IVAL *ivals)
{                               /* -- area under step function */
  FNSTEP *s, *t;                /* to traverse the function steps */
  IVAL   *v;                    /* to traverse the intervals */
  double p;                     /* left border of current rectangle */
  diff_t h;                     /* height      of current rectangle */
  double area = 0;              /* area under step function */

  assert(steps && ivals);       /* check the function arguments */
  for (s = steps, v = ivals; v->beg < INFINITY; v++) {
    p = v->beg;                 /* traverse the intervals */
    while (s->pt < v->beg) s++; /* skip all steps before interval */
    h = (--s)->ht;              /* get height before interval starts */
#if 1
    for (t = s+1; t->pt < v->end; t++) {
      area += (double)h *(t->pt -p);
      p = t->pt; h = t->ht;     /* sum the areas of the rectangles */
    }                           /* under the steps of the function */
#else
    for (t = s+1; t->pt < v->end; t++)
      if (t->ht > h) h = t->ht; /* find maximum height in interval */
#endif
    area += (double)h *(v->end -p);
  }                             /* add the area of the last rectangle */
  return area;                  /* return the computed area */
}  /* austeps() */

/*--------------------------------------------------------------------*/

static void ivlens (FNSTEP *steps, IVAL *ivals, double *lens, ITEM n)
{                               /* -- area under step function */
  FNSTEP *s, *t;                /* to traverse the function steps */
  IVAL   *v;                    /* to traverse the intervals */
  double p;                     /* left border of current rectangle */
  diff_t h;                     /* height      of current rectangle */

  assert(steps && ivals);       /* check the function arguments */
  memset(lens, 0, (size_t)(n+1) *sizeof(double));
  for (s = steps, v = ivals; v->beg < INFINITY; v++) {
    p = v->beg;                 /* traverse the intervals */
    while (s->pt < v->beg) s++; /* skip all steps before interval */
    h = (--s)->ht;              /* get height before interval starts */
    for (t = s+1; t->pt < v->end; t++) {
      lens[h] += t->pt -p;      /* sum the interval length */
      p = t->pt; h = t->ht;     /* for the current height, */
    }                           /* then go to the next point */
    lens[h] += v->end -p;       /* sum the interval length */
  }                             /* the last rectangle */
}  /* ivlens() */

/*--------------------------------------------------------------------*/

static void irs_free (IRSDATA *data)
{                               /* --- free statistics data */
  assert(data);                 /* check the function argument */
  if (data->buf)   free(data->buf);
  if (data->gls)   free(data->gls);
  if (data->steps) free(data->steps);
  free(data->ivs1);             /* delete allocated buffers */
}  /* irs_free() */

/*--------------------------------------------------------------------*/

static int irs_alloc (IRSDATA *data, TRAINSET *trnset, int stat)
{                               /* --- allocate statistics data */
  size_t n;                     /* total number of points/spikes */

  assert(data && trnset);       /* check the function arguments */
  data->trnset = trnset;        /* note the underlying train set */
  n = (size_t)tns_maxlen(trnset);
  data->ivs1  = (IVAL*)malloc((2*n+2) *sizeof(IVAL));
  if (!data->ivs1) return -1;   /* create interval list buffers */
  data->ivs2  = data->ivs1+n+1; /* and split into two buffers */
  data->steps = NULL;           /* clear step function and */
  data->gls   = NULL;           /* item-specific value buffers */
  data->buf   = NULL;
  if ((stat == STAT_CPC) || (stat == STAT_CXC)
  ||  (stat == STAT_CCF) || (stat == STAT_CCR)
  ||  (stat == STAT_CIW) || (stat == STAT_CWR)) {
    n = tns_extent(trnset);     /* get the number of points/spikes */
    data->steps = (FNSTEP*)malloc((n+n+2) *sizeof(FNSTEP));
    if (!data->steps) { irs_free(data); return -1; }
  }                             /* buffer for step function */
  if ((stat == STAT_CCF) || (stat == STAT_CCR)) {
    n = (size_t)tns_cnt(trnset);/* get the number of items */
    data->gls = (double*)malloc((n+n+2) *sizeof(double));
    if (!data->gls) { irs_free(data); return -1; }
    data->cls = data->gls +n+1; /* create complexity frequency arrays */
  }                             /* (global and conditional) */
  if ((stat == STAT_CIF) || (stat == STAT_CIR)
  ||  (stat == STAT_CIW) || (stat == STAT_CWR)) {
    n = (size_t)tns_cnt(trnset);/* get the number of items */
    data->buf = (double*)malloc(n *sizeof(double));
    if (!data->buf) { irs_free(data); return -1; }
  }                             /* buffer for item-specific values */
  return 0;                     /* return 'ok' */
}  /* irs_alloc() */

/*--------------------------------------------------------------------*/

static void irs_init (IRSDATA *data, ITEM item,
                      double width, int stat, int alpha)
{                               /* --- initialize statistics data */
  ITEM   i;                     /* to traverse the items */
  TRAIN  *t;                    /* train of item to test */
  double r;                     /* total length of point/time range */

  assert(data                   /* check the function arguments */
  &&    (item  >= 0) && (item  < tns_cnt(data->trnset))
  &&    (width >= 0) && (alpha > 0));
  r = data->range[0].end -data->range[0].beg;
  if (data->steps) {            /* if step function is needed */
    stepfn(data->steps, data->trnset, item, width);
    if ((stat == STAT_CPC) && (alpha > 1))
      powsteps(data->steps, alpha);
    if ((stat == STAT_CPC) || (stat == STAT_CXC)
    ||  (stat == STAT_CCF) || (stat == STAT_CCR)) {
      t = tns_train(data->trnset, item);
      data->len = ivs7trn(data->ivs1, data->range, t, width);
      data->avg = (r <= 0) ? 0
                : austeps(data->steps, data->range) /r;
    }                           /* compute global average for CPC */
    if ((stat == STAT_CCF) || (stat == STAT_CCR)) {
      ivlens(data->steps, data->range, data->gls,tns_cnt(data->trnset));
    }                           /* compute global interval lengths */
    if ((stat == STAT_CIW) || (stat == STAT_CWR)) {
      for (i = tns_cnt(data->trnset); --i >= 0; ) {
        if (i == item) continue;/* traverse items other than given */
        t = tns_train(data->trnset, i);
        ivs7trn(data->ivs1, data->range, t, width);
        data->buf[i] = (r <= 0) ? 0
                     : austeps(data->steps, data->ivs1) /r;
      }                         /* compute item-specific */
    } }                         /* average pattern complexities */
  else {                        /* if conditional item frequencies */
    for (i = tns_cnt(data->trnset); --i >= 0; ) {
      t = tns_train(data->trnset, i);
      data->buf[i] = (r <= 0) ? 0
                   : ivs7trn(data->ivs1, data->range, t, width) /r;
    }                           /* compute total interval length */
  }                             /* for all items */
}  /* irs_init() */

/*--------------------------------------------------------------------*/

static void irs_term (IRSDATA *data, ITEM item)
{ }                             /* --- terminate computations */

/*--------------------------------------------------------------------*/

static double irs_cpc (IRSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- cond. pattern complexities */
  double r;                     /* total length of the intervals */
  double s;                     /* value of statistic */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0));
  r = ivs7trn(data->ivs1, data->range, train, width);
  if (r <= 0) return 0;         /* compute the item intervals */
  s = austeps(data->steps, data->ivs1) /r;
  r = data->avg;                /* compute area under step function */
  return (r > 0) ? (s-r)/r : 0; /* compute the statistic value */
}  /* irs_cpc() */

/*--------------------------------------------------------------------*/

static double irs_cxc (IRSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- cond. excess complexities */
  FNSTEP *s, *t;                /* to traverse the function steps */
  IVAL   *v;                    /* to traverse the intervals */
  double r;                     /* total length of the intervals */
  double p;                     /* left border of current rectangle */
  diff_t h;                     /* height      of current rectangle */
  double a;                     /* area under step function */
  double cxc = 0;               /* value of statistic */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0));
  r = ivs7trn(data->ivs1, data->range, train, width);
  if (r <= 0) return 0;         /* compute the item intervals */
  for (s = data->steps, v = data->ivs1; v->beg < INFINITY; v++) {
    p = v->beg;                 /* traverse the intervals */
    while (s->pt < v->beg) s++; /* skip all steps before interval */
    h = (--s)->ht;              /* get height before interval starts */
    for (a = 0, t = s+1; t->pt < v->end; t++) {
      a += (double)h *(t->pt-p);/* sum the areas of the rectangles */
      p = t->pt; h = t->ht;     /* under the steps of the function */
    }                           /* that lie in the interval */
    a += (double)h *(v->end-p); /* add the area of the last rectangle */
    a = a /(v->end -v->beg) -data->avg;  /* compute excess complexity */
    if (a > 0) cxc += powi(a, alpha);
  }                             /* sum the statistic term if positive */
  return cxc /r;                /* compute the statistic value */
}  /* irs_cxc() */

/*--------------------------------------------------------------------*/

static double irs_ccf (IRSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- cond. complexity frequencies */
  ITEM   i, n;                  /* loop variable, number of items */
  double gpc;                   /* average global pattern complexity */
  double sum, r, x;             /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  r = ivs7trn(data->ivs1, data->range, train, width);
  ivlens(data->steps, data->ivs1, data->cls, tns_cnt(data->trnset));
  n = tns_cnt(data->trnset);    /* traverse the pattern complexities */
  for (gpc = 0, i = 1; i < n; i++)
    gpc += (double)i *data->gls[i];
  x = data->range[0].end -data->range[0].beg;
  gpc /= x; r /= x;             /* compute average pattern complexity */
  for (sum = 0, i = (int)ceil(gpc); i <= n; i++) {
    x = (double)data->cls[i] -r *(double)data->gls[i];
    if (x > 0) sum += x *powi((double)i -gpc, alpha);
  }                             /* sum the excess frequencies */
  return (n > 1) ? sum /(double)n : 0;
}  /* irs_ccf() */

/*--------------------------------------------------------------------*/

static double irs_ccr (IRSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- cond. complexity freq. ratio */
  ITEM   i, n;                  /* loop variable, number of items */
  double gpc;                   /* average global pattern complexity */
  double sum, r, x, y;          /* (sum of) statistic terms */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  r = ivs7trn(data->ivs1, data->range, train, width);
  ivlens(data->steps, data->ivs1, data->cls, tns_cnt(data->trnset));
  n = tns_cnt(data->trnset);    /* traverse the pattern complexities */
  for (gpc = 0, i = 1; i <= n; i++)
    gpc += (double)i *(double)data->gls[i];
  x = data->range[0].end -data->range[0].beg;
  gpc /= x; r /= x;             /* compute average pattern complexity */
  for (sum = 0, i = (int)ceil(gpc); i <= n; i++) {
    x = (double)data->cls[i] -(y = r *(double)data->gls[i]);
    if (x > 0) sum += x/(y+1) *powi((double)i -gpc, alpha);
  }                             /* sum the excess frequencies */
  return (n > 1) ? sum /(double)n : 0;
}  /* irs_ccr() */

/*--------------------------------------------------------------------*/

static double irs_cif (IRSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional item frequencies */
  ITEM   i, n;                  /* loop variables for items */
  TRAIN  *t;                    /* to traverse the trains */
  double r;                     /* total length of intervals */
  double s, cif = 0;            /* value of statistic */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  r = ivs7trn(data->ivs1, data->range, train, width);
  for (i = n = tns_cnt(data->trnset); --i >= 0; ) {
    if (i == trn_item(train))   /* traverse the items other than */
      continue;                 /* the train that is tested */
    t = tns_train(data->trnset, i);
    s = ivs7trn(data->ivs2, data->ivs1, t, width);
#if 0
    s -= r *data->buf[i];       /* compute interval intersection */
#else
    s = s/r -data->buf[i];      /* compute interval intersection */
#endif
    if (s > 0) cif += powi(s, alpha);
  }                             /* sum excess item frequencies */
  return cif / (n-1);           /* return normalized statistic */
}  /* irs_cif() */

/*--------------------------------------------------------------------*/

static double irs_cir (IRSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional item freq. ratio */
  ITEM   i, n;                  /* loop variables for items */
  TRAIN  *t;                    /* to traverse the trains */
  double r;                     /* total length of intervals */
  double s, y, cir = 0;         /* value of statistic */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  r = ivs7trn(data->ivs1, data->range, train, width);
  if (r <= 0) return 0;         /* compute the item intervals */
  for (i = n = tns_cnt(data->trnset); --i >= 0; ) {
    if (i == trn_item(train))   /* traverse the items other than */
      continue;                 /* the train that is tested */
    t = tns_train(data->trnset, i);
    s = ivs7trn(data->ivs2, data->ivs1, t, width);
    #if 0
    s -= y = r *data->buf[i];   /* compute interval intersection */
    #else
    s = s/r -(y = data->buf[i]);/* compute interval intersection */
    #endif
    if (s > 0) cir += powi(s/y, alpha);
  }                             /* sum excess item frequencies */
  return cir / (n-1);           /* return normalized statistic */
}  /* irs_cir() */

/*--------------------------------------------------------------------*/

static double irs_ciw (IRSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional item weights */
  ITEM   i, n;                  /* loop variables for items */
  TRAIN  *t;                    /* to traverse the trains */
  double r;                     /* total length of the intervals */
  double s, csw = 0;            /* value of statistic */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  r = ivs7trn(data->ivs1, data->range, train, width);
  if (r <= 0) return 0;         /* compute the item intervals */
  for (i = n = tns_cnt(data->trnset); --i >= 0; ) {
    if (i == trn_item(train))   /* traverse the items other than */
      continue;                 /* the train that is tested */
    t = tns_train(data->trnset, i);
    ivs7trn(data->ivs2, data->ivs1, t, width);
    s = austeps(data->steps, data->ivs2);
    #if 0
    s -= r *data->buf[i];       /* compute interval intersection */
    #else
    s = s/r -data->buf[i];      /* compute interval intersection */
    #endif
    if (s > 0) csw += powi(s, alpha);
  }                             /* sum excess item weights */
  return csw / (n-1);           /* return normalized statistic */
}  /* irs_ciw() */

/*--------------------------------------------------------------------*/

static double irs_cwr (IRSDATA *data, TRAIN *train,
                       double width, int alpha)
{                               /* --- conditional item weights */
  ITEM   i, n;                  /* loop variables for items */
  TRAIN  *t;                    /* to traverse the trains */
  double r;                     /* total length of the intervals */
  double s, y, cwr = 0;         /* value of statistic */

  assert(data && train          /* check the function arguments */
  &&    (width >= 0) && (alpha > 0) && (tns_cnt(data->trnset) > 1));
  r = ivs7trn(data->ivs1, data->range, train, width);
  if (r <= 0) return 0;         /* compute the item intervals */
  for (i = n = tns_cnt(data->trnset); --i >= 0; ) {
    if (i == trn_item(train))   /* traverse the items other than */
      continue;                 /* the train that is tested */
    t = tns_train(data->trnset, i);
    ivs7trn(data->ivs2, data->ivs1, t, width);
    s = austeps(data->steps, data->ivs2);
    #if 0
    s -= y = r *data->buf[i];   /* compute interval intersection */
    #else
    s = s/r -(y = data->buf[i]);/* compute interval intersection */
    #endif
    if (s > 0) cwr += powi(s/y, alpha);
  }                             /* sum excess item weights */
  return cwr / (n-1);           /* return normalized statistic */
}  /* irs_cwr() */

/*--------------------------------------------------------------------*/

static int irs_test (TRAINSET *trnset, ITEM item,
                     double width, double beg, double end,
                     int surr, RNG *rng, int stat, int alpha,
                     long runs, int qtls, FILE *out)
{                               /* --- execute statistical test */
  int       r = 0;              /* function result, error status */
  ITEM      i, n, e;            /* loop variables, buffer */
  long      k;                  /* loop variable for runs */
  size_t    z, points;          /* (total) number of points/spikes */
  TRAIN     *train, *t;         /* surrogate and original train */
  TRNSURRFN *surrfn;            /* surrogate train generation func. */
  IRSTATFN  *irsfn;             /* test statistic to compute */
  double    *stats = NULL;      /* train statistic values */
  double    *sums  = NULL;      /* sums of statistic values */
  double    x, y, total;        /* (total value of) test statistic */
  double    p;                  /* computed p-value */
  CCHAR     *s;                 /* name/identifier of an item/neuron */
  IRSDATA   data;               /* influence region statistics data */

  assert(trnset && (width >= 0) && (end >= beg) && rng
  &&    (alpha > 0) && out);    /* check the function arguments */

  /* --- initialize --- */
  train = trn_create(0, NULL, tns_maxlen(trnset));
  if (!train) return -1;        /* create a train for the surrogates */
  if (runs > 0) {               /* if to generate surrogates */
    stats = (double*)malloc((size_t)(runs +runs) *sizeof(double));
    if (!stats) { trn_delete(train); return -1; }
    sums = stats +runs;         /* allocate the statistics arrays */
    dbl_clear(sums, (size_t)runs);
  }                             /* clear the statistic sums */
  total = 0; points = 0;        /* initialize the statistic total */
  if (irs_alloc(&data, trnset, stat) != 0) {
    if (stats) { free(stats); } trn_delete(train); return -1; }
  if      (surr == 'r')      surrfn = trn_random;
  else if (surr == 'd')      surrfn = trn_dither;
  else if (surr == 's')      surrfn = trn_shift;
  else                       surrfn = trn_random;
  if      (stat == STAT_CPC) irsfn  = irs_cpc;
  else if (stat == STAT_CXC) irsfn  = irs_cxc;
  else if (stat == STAT_CCF) irsfn  = irs_ccf;
  else if (stat == STAT_CCR) irsfn  = irs_ccr;
  else if (stat == STAT_CIF) irsfn  = irs_cif;
  else if (stat == STAT_CIR) irsfn  = irs_cir;
  else if (stat == STAT_CIW) irsfn  = irs_ciw;
  else if (stat == STAT_CWR) irsfn  = irs_cwr;
  else                       irsfn  = irs_cpc;
  data.range[0].beg = beg;      /* initialize the point/time range */
  data.range[0].end = end;
  data.range[1].beg = INFINITY; /* store a sentinel at the end */
  data.range[1].end = INFINITY;
  header(out, runs, qtls, " %11s");

  /* --- execute tests for items --- */
  n = tns_cnt(trnset);          /* get the number of trains */
  if (item >= 0) e = (i = item)+1;
  else         { e = n; i = 0; }/* get the range of items */
  for ( ; i < e; i++) {         /* traverse the items to analyze */
    t = tns_train(trnset, i);   /* sum the number of points */
    points += z = (size_t)trn_len(t);
    irs_init(&data, i, width, stat, alpha);
    x = (z > 0) ? irsfn(&data, t, width, alpha) : 0;
    total += x;                 /* compute and sum the test statistic */
    s = ib_name(tns_base(trnset), i);
    if (runs <= 0) {            /* if no surrogate data requested, */
      fprintf(out, "%6s  %11.6f", s, x); /* report statistic value */
      if (qtls) fprintf(out, " %6"SIZE_FMT, z); }
    else {                      /* if to compare to surrogate data */
      for (p = 0, k = runs; --k >= 0; ) {
        train = surrfn(tns_train(trnset, i), beg, end, rng, train);
        trn_sort(train, +1);    /* create a surrogate train */
        y = (z > 0) ? irsfn(&data, train, width, alpha) : 0;
        stats[k] = y;           /* compute and store test statistic */
        sums[k] += y;           /* sum test statistic over items */
        if (y >= x) p++;        /* count the values exceeding or */
      }                         /* meeting the actual statistic value */
      p /= (double)runs;        /* compute and print the p-value */
      fprintf(out, "%6s  %8.6f%c", s, p, sigind(p));
      if (qtls) {               /* if extended output requested */
        dbl_qsort(stats, (size_t)runs, +1);
        report(out, " %11.6f", stats, runs, x, z);
      }                         /* sort statistic values and */
    }                           /* report quantiles of statistic */
    fprintf(out, "\n");         /* and the actual statistic value, */
    irs_term(&data, i);         /* then terminate the output line */
  }                             /* and the statistic computation */

  /* --- total statistics --- */
  if ((runs > 0) && (item < 0)) /* if all neurons were checked */
    footer(out, sums, runs, total, points, qtls);

  /* --- clean up --- */
  irs_free(&data);              /* delete the work memory, */
  if (stats) free(stats);       /* the statistics arrays */
  trn_delete(train);            /* and the surrogate train */
  return r;                     /* return the error status */
}  /* irs_test() */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/
#ifdef NAPA_MAIN

static void help()
{                               /* --- print add. option information */
  #ifndef QUIET
  fprintf(stderr, "\n");        /* terminate startup message */
  printf("statistic computation modes (option -m#)\n");
  printf("  b     time bins           (discretized time)\n");
  printf("  p     point-centered bins (continuous  time)\n");
  printf("  w     covering windows    (continuous  time)\n");
  printf("  i     influence regions   (continuous  time)\n");
  printf("test statistics (option -s#\n");
  printf("  cpc   conditional pattern complexity\n");
  printf("  cxc   conditional excess complexity\n");
  printf("  ccf   conditional complexity frequency\n");
  printf("  ccr   conditional complexity frequency ratio\n");
  printf("  cif   conditional item frequency\n");
  printf("  cir   conditional item frequency ratio\n");
  printf("  ciw   conditional item weight\n");
  printf("  cwr   conditional item weight ratio\n");
  printf("  cpo   conditional pattern overlap\n");
  printf("  bre   background rate estimation\n");
  printf("Test statistic bre is available only with option -mb,\n"
         "test statistic cpo only with options -mb and -mp.\n");
  printf("surrogate data generation method (option -g#)\n");
  printf("  r     point/spike time randomization\n");
  printf("  d     point/spike dithering/displacement\n");
  printf("  s     train shifting/dithering\n");
  printf("random function type for displacements (option -d#)\n");
  printf("  u/r   uniform/rectangular density\n");
  printf("  t     symmetric triangular density\n");
  printf("  g/n   Gaussian/normal density\n");
  #endif                        /* print help information */
  exit(0);                      /* abort the program */
}  /* help() */

/*--------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (rng)    rng_delete(rng);       \
  if (trnset) tns_delete(trnset, 0); \
  if (tread)  trd_delete(tread,  1); \
  if (ibase)  ib_delete (ibase);     \
  if (out && (out != stdout)) fclose(out);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counters */
  char    *s;                   /* to traverse the options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_inp  = NULL;      /* name of input  file */
  CCHAR   *fn_out  = NULL;      /* name of output file */
  CCHAR   *fn_sel  = NULL;      /* name of item selection file */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *comment = NULL;      /* comment characters */
  CCHAR   *iname   = NULL;      /* name/identifier of item to test */
  CCHAR   *sname   = "cpc";     /* name of test statistic */
  double  width    = 0.003;     /* width of window/influence region */
  double  beg      = NAN;       /* minimum of point/time range */
  double  end      = NAN;       /* maximum of point/time range */
  int     mode     = 'w';       /* identifier of statistics mode */
  int     stat     = STAT_CPC;  /* identifier of test statistic */
  int     alpha    = 1;         /* weighting exponent for statistic */
  long    runs     = 1000;      /* number of surrogate runs */
  int     surr     = 'r';       /* surrogate data generation method */
  int     prob     = 'u';       /* probability density function type */
  double  sigma    = 0.005;     /* dispersion parameter */
  int     qtls     = 0;         /* flag for quantiles output */
  int     minp     = TRN_ITEMPT;/* mode for train set reading */
  long    seed;                 /* seed value for random numbers */
  ITEM    m, item;              /* number of items/trains, item */
  size_t  n;                    /* number of points/spikes */
  clock_t t;                    /* timer for measurements */

  #ifndef QUIET                 /* if not quiet version */
  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no arguments given */
    printf("usage: %s [options] infile [outfile]\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-i#      name/identifier of item to check         "
                    "(default: check all)\n");
    printf("-w#      width of time window/influence region    "
                    "(default: %g)\n", width);
    printf("-a#      start of point/time range                "
                    "(default: auto)\n");
    printf("-z#      end   of point/time range                "
                    "(default: auto)\n");
    printf("-m#      statistic computation mode               "
                    "(default: w)\n");
    printf("-s#      test statistic to use                    "
                    "(default: %s)\n", sname);
    printf("-x#      weighting exponent for test statistic    "
                    "(default: %d)\n", alpha);
    printf("-n#      number of surrogate trains to generate   "
                    "(default: %ld)\n", runs);
    printf("-g#      surrogate data generation method         "
                    "(default: %c)\n", surr);
    printf("-d#      random function type for displacements   "
                    "(default: %c)\n", prob);
    printf("-p#      parameter for random function density    "
                    "(default: %g)\n", sigma);
    printf("         (standard deviation or half base width "
                    "of rectangle/triangle)\n");
    printf("-S#      seed for pseudo-random numbers           "
                    "(default: time)\n");
    printf("-q       extended output (print quantiles)        "
                    "(default: only p-values)\n");
    printf("-l       one train per record (item + points)     "
                    "(default: id/time pairs)\n");
    printf("-y       no item (with -l) or item after point    "
                    "(default: item first)\n");
    printf("-R#      read an item selection from a file\n");
    printf("-r#      record/transaction separators            "
                    "(default: \"\\n\")\n");
    printf("-f#      field /item        separators            "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                       "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                       "
                    "(default: \"#\")\n");
    printf("-!       print additional option information\n");
    printf("infile   file to read trains from                 "
                    "[required]\n");
    printf("outfile  file to write computed statistics to     "
                    "[optional]\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */
  #endif  /* #ifndef QUIET */

  /* --- evaluate arguments --- */
  seed = (long)time(NULL);      /* get the default seed value */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (*s) {              /* traverse options */
        switch (*s++) {         /* evaluate switches */
          case '!': help();                         break;
          case 'i': optarg = &iname;                break;
          case 'w': width  =      strtod(s, &s);    break;
          case 'a': beg    =      strtod(s, &s);    break;
          case 'z': end    =      strtod(s, &s);    break;
          case 'm': mode   = (*s) ? *s++ : 'w';     break;
          case 's': optarg = &sname;                break;
          case 'x': alpha  = (int)strtol(s, &s, 0); break;
          case 'n': runs   =      strtol(s, &s, 0); break;
          case 'g': surr   = (*s) ? *s++ : 'r';     break;
          case 'd': prob   = (*s) ? *s++ : 'u';     break;
          case 'p': sigma  =      strtod(s, &s);    break;
          case 'S': seed   =      strtol(s, &s, 0); break;
          case 'q': qtls   = -1;                    break;
          case 'l': minp  &= ~TRN_ITEMPT;           break;
          case 'y': minp  |=  TRN_PURETRN;          break;
          case 'R': optarg = &fn_sel;               break;
          case 'r': optarg = &recseps;              break;
          case 'f': optarg = &fldseps;              break;
          case 'b': optarg = &blanks;               break;
          case 'C': optarg = &comment;              break;
          default : error(E_OPTION, *--s);          break;
        }                       /* set option variables */
        if (optarg && *s) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-options */
        case  0: fn_inp = s;      break;
        case  1: fn_out = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg)     error(E_OPTARG);    /* check (option) arguments */
  if (k < 1)      error(E_ARGCNT);    /* and number of arguments */
  if (width <= 0) error(E_WIDTH, width);
  if (alpha <= 0) error(E_ALPHA, alpha);
  if (strchr("bpwi", mode) == NULL)
    error(E_MODE, mode);        /* check the statistic mode */
  if      (strcmp(sname, "cpc") == 0) stat = STAT_CPC;
  else if (strcmp(sname, "cxc") == 0) stat = STAT_CXC;
  else if (strcmp(sname, "ccf") == 0) stat = STAT_CCF;
  else if (strcmp(sname, "ccr") == 0) stat = STAT_CCR;
  else if (strcmp(sname, "cif") == 0) stat = STAT_CIF;
  else if (strcmp(sname, "cir") == 0) stat = STAT_CIR;
  else if (strcmp(sname, "ciw") == 0) stat = STAT_CIW;
  else if (strcmp(sname, "cwr") == 0) stat = STAT_CWR;
  else if (strcmp(sname, "cpo") == 0) stat = STAT_CPO;
  else if (strcmp(sname, "bre") == 0) stat = STAT_BRE;
  else error(E_STAT, sname);    /* decode the statistic name */
  if (((stat == STAT_BRE) && (mode != 'b'))
  ||  ((stat == STAT_CPO) && (strchr("bp", mode) == NULL)))
    error(E_STAT, sname);       /* bre and cpo only for bins */
  if (strchr("rds", surr) == NULL)
    error(E_SURR, surr);        /* check surrogate generation method */
  switch (prob) {               /* evaluate prob. density character */
    case 'u': prob = RNG_UNIFORM;    break;
    case 'r': prob = RNG_RECT;       break;
    case 't': prob = RNG_TRIANG;     break;
    case 'z': prob = RNG_GAUSS;      break;
    case 'n': prob = RNG_NORMAL;     break;
    default : error(E_RANDOM, prob); break;
  }                             /* get the prob. density identifier */
  if (sigma < 0) error(E_SIGMA, sigma);
  MSG(stderr, "\n");            /* terminate the startup message */

  /* --- read item selection --- */
  ibase = ib_create(0, 0);      /* create an item base */
  if (!ibase) error(E_NOMEM);   /* to manage the items */
  tread = trd_create();         /* create a transaction reader */
  if (!tread) error(E_NOMEM);   /* and configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, "", comment);
  if (fn_sel) {                 /* if item appearances are given */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_sel) != 0)
      error(E_FOPEN, trd_name(tread));
    MSG(stderr, "reading %s ... ", trd_name(tread));
    m = ib_readsel(ibase,tread);/* read the given item selection */
    if (m < 0) error((int)-m, ib_errmsg(ibase, NULL, 0));
    trd_close(tread);           /* close the input file */
    MSG(stderr, "[%"ITEM_FMT" item(s)]", m);
    MSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }                             /* print a log message */

  /* --- read train set --- */
  trnset = tns_create(ibase);   /* create a train set */
  if (!trnset) error(E_NOMEM);  /* to store the trains */
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_inp) != 0)
    error(E_FOPEN, trd_name(tread));
  MSG(stderr, "reading %s ... ", trd_name(tread));
  k = tns_read(trnset, tread, minp);
  if (k < 0)                    /* read the transaction database */
    error(-k, tns_errmsg(trnset, NULL, 0));
  trd_delete(tread, 1);         /* close the input file and */
  tread = NULL;                 /* delete the table reader */
  m = tns_cnt(trnset);          /* get the number of trains */
  n = tns_extent(trnset);       /* and the number of points */
  MSG(stderr, "[%"ITEM_FMT" train(s), %"SIZE_FMT" point(s)]", m, n);
  MSG(stderr, " done [%.2fs].", SEC_SINCE(t));
  if ((m < 2) || (n <= 0))      /* check for at least two items */
    error(E_ITEMS);             /* and at least one point/spike */
  MSG(stderr, "\n");            /* terminate the log message */
  item = (iname) ? ib_item(tns_base(trnset), iname) : -1;
  if (isnan(beg)) beg = floor(tns_min(trnset, 1));
  if (isnan(end)) end = ceil (tns_max(trnset, 1));

  /* --- sort train set --- */
  t = clock();                  /* start timer, print log message */
  MSG(stderr, "filtering and sorting points ... ");
  tns_filter(trnset, beg, end); /* filter points with region */
  tns_sort(trnset, +1);         /* sort the trains by points */
  MSG(stderr, "[%"ITEM_FMT" train(s), %"SIZE_FMT" point(s)]", m, n);
  MSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- do statistical tests --- */
  rng = rng_createx((unsigned)seed, prob, sigma);
  if (!rng) error(E_NOMEM);     /* create a random number generator */
  if (fn_out && (strcmp(fn_out, "-") == 0)) fn_out = "";
  if (fn_out && *fn_out) { out = fopen(fn_out, "w"); }
  else                   { out = stdout; fn_out = "<stdout>"; }
  t = clock();                  /* start timer, print log message */
  fprintf(stderr, "writing %s ... ", fn_out);
  if (!out) error(E_FOPEN, fn_out);
  if      (mode == 'b')         /* if test based on time bins */
    k = bns_test(trnset, item, width, beg, end,
                 surr, rng, stat, alpha, runs, qtls, out);
  else if (mode == 'p')         /* if test based on point bins */
    k = pbs_test(trnset, item, width, beg, end,
                 surr, rng, stat, alpha, runs, qtls, out);
  else if (mode == 'w')         /* if test based on covering windows */
    k = cws_test(trnset, item, width, beg, end,
                 surr, rng, stat, alpha, runs, qtls, out);
  else if (mode == 'i')         /* if test based on influence regions */
    k = irs_test(trnset, item, width, beg, end,
                 surr, rng, stat, alpha, runs, qtls, out);
  if (k) error(E_NOMEM);        /* execute the statistical tests */
  if (((out == stdout) ? fflush(out) : fclose(out)) != 0)
    error(E_FWRITE, fn_out);    /* close the output file and */
  out = NULL;                   /* print a success message */
  MSG(stderr, "[%"ITEM_FMT" train(s), %"SIZE_FMT" point(s)]", m, n);
  MSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */

#endif
